"""Scout scan to Autopilot Plan builder."""

from __future__ import annotations

import re
from typing import Any, Iterable

from g_gremlin.scout.models import EvidenceRow, RecommendedAction, ScoutFinding, ScoutScanResult

from .contracts import PlanRisk, PlanSource, PlanStep, PlanV1
from .tools import AutopilotToolDefinition, ToolRegistry


_RISK_LEVEL_MAP = {
    "safe": 0,
    "low": 1,
    "medium": 2,
    "high": 3,
}

_BLAST_ORDER = {
    "single_record": 0,
    "bulk_records": 1,
    "metadata": 2,
    "external_side_effect": 3,
}

_SFID_RE = re.compile(r"[A-Za-z0-9]{15,18}")
_ACTION_ALIASES = {
    "create_task": "tasks.create",
    "lead_to_account_match_health": "scout.lead_to_account_match_health",
    "deactivate_user": "sfdc.user.deactivate",
    "reassign_owner": "sfdc.record.reassign_owner",
    "dealdesk_renewal_create": "dealdesk.renewal.create_shell",
    "dealdesk_stamp": "dealdesk.approval.stamp",
    "opportunity_field_update": "sfdc.record.update",
}


def _action_risk_level(action: RecommendedAction, tool_def: AutopilotToolDefinition | None) -> int:
    if tool_def is None:
        return 3
    if tool_def.risk_level is not None:
        return int(tool_def.risk_level)
    raw = action.risk_level if action and action.risk_level is not None else ""
    if isinstance(raw, (int, float)):
        return max(0, min(3, int(raw)))
    if isinstance(raw, str):
        return _RISK_LEVEL_MAP.get(raw.strip().lower(), 3)
    return 3


def _build_evidence_payload(rows: Iterable[EvidenceRow]) -> list[dict[str, Any]]:
    payload = []
    for row in rows:
        payload.append({
            "record_id": row.record_id,
            "sobject": row.sobject,
            "fields": row.fields,
            "reason": row.reason,
            "reason_code": row.reason_code,
            "severity": row.severity,
            "related_ids": list(row.related_ids),
        })
    return payload


def _build_step_args(
    finding: ScoutFinding,
    action: RecommendedAction,
    evidence_rows: list[EvidenceRow],
    tool_name: str,
) -> dict[str, Any]:
    record_ids = [row.record_id for row in evidence_rows if row.record_id]
    payload = {
        "scout_name": finding.scout_name,
        "action_type": action.action_type,
        "action_details": action.details,
        "record_ids": record_ids,
        "records": _build_evidence_payload(evidence_rows),
    }

    if tool_name == "sfdc.user.assign_permission_sets":
        user_ids = _extract_user_ids(evidence_rows, record_ids)
        permission_set_ids = _extract_permission_set_ids(finding, action, evidence_rows)
        payload["user_ids"] = user_ids
        payload["permission_set_ids"] = permission_set_ids

    if tool_name == "sfdc.user.assign_permission_set_groups":
        user_ids = _extract_user_ids(evidence_rows, record_ids)
        group_ids = _extract_permission_set_group_ids(finding, action, evidence_rows)
        payload["user_ids"] = user_ids
        payload["permission_set_group_ids"] = group_ids

    if tool_name == "sfdc.user.add_to_queue_group":
        user_ids = _extract_user_ids(evidence_rows, record_ids)
        group_ids = _extract_queue_group_ids(finding, action, evidence_rows)
        payload["user_ids"] = user_ids
        payload["group_ids"] = group_ids

    if tool_name == "sfdc.forecast.assign_user_to_role":
        user_ids = _extract_user_ids(evidence_rows, record_ids)
        role_ids = _extract_forecast_role_ids(finding, action, evidence_rows)
        payload["user_ids"] = user_ids
        payload["role_ids"] = role_ids
        if isinstance(action.details, dict) and action.details.get("assignments"):
            payload["assignments"] = action.details.get("assignments")

    if tool_name == "tasks.create":
        owner_id = _extract_owner_id(finding, evidence_rows)
        if owner_id:
            payload["owner_id"] = owner_id
        payload["subject"] = _render_task_subject(finding, action)
        payload["description"] = _render_task_description(finding, action, evidence_rows)
        due_date = _extract_due_date(finding, action, evidence_rows)
        if due_date:
            payload["due_date"] = due_date

    if tool_name == "sfdc.record.update":
        if evidence_rows:
            payload["record_id"] = evidence_rows[0].record_id
            details = action.details or {}
            if isinstance(details, dict) and details.get("sobject"):
                payload["sobject"] = details.get("sobject")
            elif evidence_rows[0].sobject:
                payload["sobject"] = evidence_rows[0].sobject
            else:
                payload["sobject"] = "Opportunity"
            row_fields = evidence_rows[0].fields if evidence_rows[0].fields else {}
            payload["fields"] = row_fields.get("proposed_diffs", {})

    return payload


def _as_list(value: Any) -> list[str]:
    if value is None:
        return []
    if isinstance(value, list):
        return [str(item) for item in value if item not in (None, "")]
    return [str(value)]


def _dedupe(values: list[str]) -> list[str]:
    seen = set()
    ordered: list[str] = []
    for value in values:
        if value in seen:
            continue
        seen.add(value)
        ordered.append(value)
    return ordered


def _extract_permission_set_ids(
    finding: ScoutFinding,
    action: RecommendedAction,
    evidence_rows: list[EvidenceRow],
) -> list[str]:
    ids: list[str] = []
    details = action.details or {}
    if isinstance(details, dict):
        ids.extend(_as_list(details.get("permission_set_ids")))
        ids.extend(_as_list(details.get("permission_set_id")))
        ids.extend(_as_list(details.get("permission_set_names")))

    params_used = finding.parameters_used or {}
    if isinstance(params_used, dict):
        ids.extend(_as_list(params_used.get("permission_set_ids")))
        ids.extend(_as_list(params_used.get("permission_set_id")))
        ids.extend(_as_list(params_used.get("permission_set_names")))

    for row in evidence_rows:
        fields = row.fields or {}
        if isinstance(fields, dict):
            for key in ("permission_set_id", "PermissionSetId", "permissionSetId"):
                if key in fields:
                    ids.extend(_as_list(fields.get(key)))
        reason_code = row.reason_code or ""
        if "permission" in reason_code.lower():
            ids.extend(_SFID_RE.findall(reason_code))

    return _dedupe(ids)


def _extract_user_ids(evidence_rows: list[EvidenceRow], record_ids: list[str]) -> list[str]:
    user_ids = [
        row.record_id
        for row in evidence_rows
        if row.record_id and (row.sobject or "").lower() == "user"
    ]
    if not user_ids:
        return record_ids
    return user_ids


def _extract_permission_set_group_ids(
    finding: ScoutFinding,
    action: RecommendedAction,
    evidence_rows: list[EvidenceRow],
) -> list[str]:
    ids: list[str] = []
    details = action.details or {}
    if isinstance(details, dict):
        ids.extend(_as_list(details.get("permission_set_group_ids")))
        ids.extend(_as_list(details.get("permission_set_group_id")))
        ids.extend(_as_list(details.get("permission_set_group_names")))

    params_used = finding.parameters_used or {}
    if isinstance(params_used, dict):
        ids.extend(_as_list(params_used.get("permission_set_group_ids")))
        ids.extend(_as_list(params_used.get("permission_set_group_id")))
        ids.extend(_as_list(params_used.get("permission_set_group_names")))

    for row in evidence_rows:
        fields = row.fields or {}
        if isinstance(fields, dict):
            for key in ("PermissionSetGroupId", "permission_set_group_id"):
                if key in fields:
                    ids.extend(_as_list(fields.get(key)))
        reason_code = row.reason_code or ""
        if "group" in reason_code.lower():
            ids.extend(_SFID_RE.findall(reason_code))

    return _dedupe(ids)


def _extract_queue_group_ids(
    finding: ScoutFinding,
    action: RecommendedAction,
    evidence_rows: list[EvidenceRow],
) -> list[str]:
    ids: list[str] = []
    details = action.details or {}
    if isinstance(details, dict):
        ids.extend(_as_list(details.get("group_ids")))
        ids.extend(_as_list(details.get("group_id")))
        ids.extend(_as_list(details.get("queue_group_ids")))
        ids.extend(_as_list(details.get("queue_group_id")))
        ids.extend(_as_list(details.get("group_names")))
        ids.extend(_as_list(details.get("queue_names")))

    params_used = finding.parameters_used or {}
    if isinstance(params_used, dict):
        ids.extend(_as_list(params_used.get("group_ids")))
        ids.extend(_as_list(params_used.get("group_id")))
        ids.extend(_as_list(params_used.get("queue_group_ids")))
        ids.extend(_as_list(params_used.get("queue_group_id")))
        ids.extend(_as_list(params_used.get("group_names")))
        ids.extend(_as_list(params_used.get("queue_names")))

    for row in evidence_rows:
        fields = row.fields or {}
        if isinstance(fields, dict):
            for key in ("GroupId", "QueueId", "group_id", "queue_id"):
                if key in fields:
                    ids.extend(_as_list(fields.get(key)))
        reason_code = row.reason_code or ""
        if "queue" in reason_code.lower() or "group" in reason_code.lower():
            ids.extend(_SFID_RE.findall(reason_code))

    return _dedupe(ids)


def _extract_forecast_role_ids(
    finding: ScoutFinding,
    action: RecommendedAction,
    evidence_rows: list[EvidenceRow],
) -> list[str]:
    ids: list[str] = []
    details = action.details or {}
    if isinstance(details, dict):
        ids.extend(_as_list(details.get("role_ids")))
        ids.extend(_as_list(details.get("role_id")))
        ids.extend(_as_list(details.get("forecast_role_ids")))
        ids.extend(_as_list(details.get("forecast_role_id")))

    params_used = finding.parameters_used or {}
    if isinstance(params_used, dict):
        ids.extend(_as_list(params_used.get("role_ids")))
        ids.extend(_as_list(params_used.get("role_id")))
        ids.extend(_as_list(params_used.get("forecast_role_ids")))
        ids.extend(_as_list(params_used.get("forecast_role_id")))

    for row in evidence_rows:
        fields = row.fields or {}
        if isinstance(fields, dict):
            for key in ("ForecastRoleId", "forecast_role_id", "RoleId", "role_id"):
                if key in fields:
                    ids.extend(_as_list(fields.get(key)))
        reason_code = row.reason_code or ""
        if "forecast" in reason_code.lower() or "role" in reason_code.lower():
            ids.extend(_SFID_RE.findall(reason_code))

    return _dedupe(ids)


def _resolve_tool_name(action_type: str) -> str:
    return _ACTION_ALIASES.get(action_type, action_type)


def _render_task_subject(finding: ScoutFinding, action: RecommendedAction) -> str:
    details = action.details or {}
    if isinstance(details, dict) and details.get("subject"):
        return str(details["subject"])
    name = finding.display_name or finding.scout_name
    return f"Autopilot follow-up: {name}"


def _render_task_description(
    finding: ScoutFinding,
    action: RecommendedAction,
    evidence_rows: list[EvidenceRow],
) -> str:
    details = action.details or {}
    if isinstance(details, dict) and details.get("description"):
        return str(details["description"])
    if evidence_rows:
        row = evidence_rows[0]
        if row.reason:
            return str(row.reason)
        if row.reason_code:
            return f"Reason: {row.reason_code}"
    return f"Generated from scout {finding.scout_name}"


def _extract_owner_id(finding: ScoutFinding, evidence_rows: list[EvidenceRow]) -> str | None:
    details = finding.parameters_used or {}
    if isinstance(details, dict):
        owner = (
            details.get("owner_id")
            or details.get("ownerId")
            or details.get("OwnerId")
            or details.get("task_owner_id")
        )
        if owner:
            return str(owner)
    for row in evidence_rows:
        fields = row.fields or {}
        if isinstance(fields, dict):
            owner = fields.get("OwnerId") or fields.get("owner_id") or fields.get("ownerid")
            if owner:
                return str(owner)
    return None


def _extract_due_date(
    finding: ScoutFinding,
    action: RecommendedAction,
    evidence_rows: list[EvidenceRow],
) -> str | None:
    details = action.details or {}
    if isinstance(details, dict):
        due = details.get("due_date") or details.get("dueDate") or details.get("ActivityDate")
        if due:
            return str(due)
    for row in evidence_rows:
        fields = row.fields or {}
        if isinstance(fields, dict):
            due = fields.get("ActivityDate") or fields.get("due_date") or fields.get("DueDate")
            if due:
                return str(due)
    return None


def _best_effort_diff_preview(evidence_rows: list[EvidenceRow]) -> dict[str, Any] | None:
    if not evidence_rows:
        return None
    row = evidence_rows[0]
    if not row.fields:
        return None
    return {
        "record_id": row.record_id,
        "sobject": row.sobject,
        "fields": row.fields,
        "reason_code": row.reason_code,
    }


def _build_step(
    *,
    step_id: str,
    run_id: str,
    finding: ScoutFinding,
    action: RecommendedAction,
    tool_name: str,
    tool_def: AutopilotToolDefinition | None,
) -> PlanStep:
    evidence_rows = list(finding.evidence)
    risk_level = _action_risk_level(action, tool_def)
    args = _build_step_args(finding, action, evidence_rows, tool_name)
    return PlanStep(
        step_id=step_id,
        tool=tool_name,
        diff_preview=_best_effort_diff_preview(evidence_rows),
        args=args,
        risk_level=risk_level,
        idempotency_key=f"{run_id}:{step_id}",
        prechecks=list(tool_def.default_prechecks) if tool_def else [],
        postchecks=list(tool_def.default_postchecks) if tool_def else [],
        evidence_ids=[f"{finding.scout_name}:{idx}" for idx, _ in enumerate(evidence_rows)],
    )


def _resolve_reassignment_config(
    finding: ScoutFinding,
    action: RecommendedAction,
) -> dict[str, Any]:
    details = action.details or {}
    if isinstance(details, dict) and isinstance(details.get("reassignment_config"), dict):
        return details.get("reassignment_config") or {}
    params_used = finding.parameters_used or {}
    if isinstance(params_used, dict) and isinstance(params_used.get("reassignment_config"), dict):
        return params_used.get("reassignment_config") or {}
    return {}


def _build_reassign_steps(
    *,
    run_id: str,
    step_index: int,
    finding: ScoutFinding,
    action: RecommendedAction,
    tool_name: str,
    tool_def: AutopilotToolDefinition | None,
) -> tuple[list[PlanStep], int]:
    config = _resolve_reassignment_config(finding, action)
    if not config:
        return [], step_index
    config_map = {str(k).lower(): v for k, v in config.items() if isinstance(v, dict)}
    grouped: dict[str, dict[str, Any]] = {}
    for row in finding.evidence:
        if not row.sobject:
            continue
        key = str(row.sobject).lower()
        cfg = config_map.get(key)
        if not isinstance(cfg, dict):
            continue
        new_owner_id = cfg.get("default_owner_id") or cfg.get("default_queue_id")
        if not new_owner_id:
            continue
        group = grouped.setdefault(
            key,
            {"sobject": row.sobject, "new_owner_id": new_owner_id, "rows": []},
        )
        group["rows"].append(row)

    steps: list[PlanStep] = []
    for group in grouped.values():
        evidence_rows = list(group["rows"])
        risk_level = _action_risk_level(action, tool_def)
        args = _build_step_args(finding, action, evidence_rows, tool_name)
        args["sobject"] = group["sobject"]
        args["new_owner_id"] = group["new_owner_id"]
        step_id = f"s{step_index}"
        step_index += 1
        steps.append(
            PlanStep(
                step_id=step_id,
                tool=tool_name,
                diff_preview=_best_effort_diff_preview(evidence_rows),
                args=args,
                risk_level=risk_level,
                idempotency_key=f"{run_id}:{step_id}",
                prechecks=list(tool_def.default_prechecks) if tool_def else [],
                postchecks=list(tool_def.default_postchecks) if tool_def else [],
                evidence_ids=[f"{finding.scout_name}:{idx}" for idx, _ in enumerate(evidence_rows)],
            )
        )
    return steps, step_index


def _summarize_plan(findings: list[ScoutFinding], steps: list[PlanStep]) -> str:
    if not steps:
        return "No executable actions found in scout scan."
    actionable_findings = [f for f in findings if f.recommended_actions]
    if len(actionable_findings) == 1:
        finding = actionable_findings[0]
        name = finding.display_name or finding.scout_name
        return f"{name}: {finding.findings_count} finding(s) -> {len(steps)} step(s)"
    return f"{len(steps)} step(s) from {len(actionable_findings)} finding(s)"


def _compute_plan_risk(
    steps: list[PlanStep],
    tool_defs: list[AutopilotToolDefinition | None],
    evidence_rows: list[EvidenceRow],
) -> PlanRisk:
    max_risk = 0
    irreversible = False
    blast_radius = "single_record"

    for step, tool_def in zip(steps, tool_defs):
        max_risk = max(max_risk, step.risk_level)
        if tool_def and tool_def.risk_category in ("irreversible", "metadata_write"):
            irreversible = True
        if step.risk_level >= 3:
            irreversible = True
        radius = tool_def.blast_radius if tool_def else None
        if radius and _BLAST_ORDER.get(radius, 0) > _BLAST_ORDER.get(blast_radius, 0):
            blast_radius = radius

    estimated_count = len([row for row in evidence_rows if row.record_id])
    target_object = None
    sobjects = {row.sobject for row in evidence_rows if row.sobject}
    if len(sobjects) == 1:
        target_object = next(iter(sobjects))

    return PlanRisk(
        max_risk_level=max_risk,
        irreversible=irreversible,
        blast_radius=blast_radius,
        estimated_count=estimated_count or None,
        target_object=target_object,
    )


def build_plan(scan_result: ScoutScanResult, registry: ToolRegistry) -> PlanV1:
    run_id = scan_result.run_id
    steps: list[PlanStep] = []
    step_tool_defs: list[AutopilotToolDefinition | None] = []
    evidence_rows: list[EvidenceRow] = []
    verification_targets: list[str] = []
    verification_seen: set[str] = set()
    step_index = 1

    for finding in scan_result.findings:
        if not finding.recommended_actions:
            continue
        if finding.scout_name and finding.scout_name not in verification_seen:
            verification_seen.add(finding.scout_name)
            verification_targets.append(finding.scout_name)
        evidence_rows.extend(list(finding.evidence))
        for action in finding.recommended_actions:
            tool_name = _resolve_tool_name(action.action_type)
            tool_def = registry.get_tool(tool_name)
            if tool_name == "sfdc.record.reassign_owner":
                new_steps, step_index = _build_reassign_steps(
                    run_id=run_id,
                    step_index=step_index,
                    finding=finding,
                    action=action,
                    tool_name=tool_name,
                    tool_def=tool_def,
                )
                for step in new_steps:
                    steps.append(step)
                    step_tool_defs.append(tool_def)
                continue
            step_id = f"s{step_index}"
            step_index += 1
            steps.append(
                _build_step(
                    step_id=step_id,
                    run_id=run_id,
                    finding=finding,
                    action=action,
                    tool_name=tool_name,
                    tool_def=tool_def,
                )
            )
            step_tool_defs.append(tool_def)

    summary = _summarize_plan(scan_result.findings, steps)
    risk = _compute_plan_risk(steps, step_tool_defs, evidence_rows)
    approvals_required = risk.max_risk_level >= 1

    plan = PlanV1(
        version=1,
        run_id=run_id,
        source=PlanSource(type="scout_scan", scan_id=scan_result.run_id),
        summary=summary,
        risk=risk,
        plan_hash=None,
        policy_hash=None,
        steps=steps,
        verification=verification_targets,
        approvals_required=approvals_required,
    )
    return plan


__all__ = ["build_plan"]



"""Gremlin v2 formula fixer API."""

from __future__ import annotations

import json
import logging
from typing import Any, List, Optional
from uuid import uuid4

from fastapi import APIRouter, Depends, HTTPException, status
from pydantic import BaseModel, Field

from ... import settings as saas_settings
from ...auth_security import require_account
from ..progress import get_redis
from ...services import formula_fixer_service as fixer_service
from g_gremlin import graph as gremlin_graph

logger = logging.getLogger(__name__)

router = APIRouter(prefix="/api/v2/gremlin/formula-fixer", tags=["gremlin"])


class SparseCell(BaseModel):
    a1: str
    formula: Optional[str] = None
    display_value: Optional[Any] = None
    sheet: str


class FormulaFixerScanRequest(BaseModel):
    spreadsheet_id: str
    sheet_name: Optional[str] = None
    scan_mode: str = Field("current_tab", description="current_tab|selected_tabs|all_tabs")
    selected_sheets: Optional[List[str]] = None
    sheet_titles: Optional[List[str]] = None
    cells: List[SparseCell]


class HealthModel(BaseModel):
    score: int
    grade: str


class FixProposal(BaseModel):
    cell: str
    expected_original: Optional[str] = None
    proposed: str


class IssueGroup(BaseModel):
    issue_id: Optional[str] = None
    type: str
    title: str = ""
    cells: List[str]
    sample_original: Optional[str] = None
    sample_fixed: Optional[str] = None
    fix_available: bool = False
    fix_confidence_band: str
    recommended_action: str
    diagnostic_actions: List[str] = Field(default_factory=list)
    fixes: List[FixProposal] = Field(default_factory=list)
    explanation: str = ""


class FormulaFixerScanResponse(BaseModel):
    health: HealthModel
    issue_groups: List[IssueGroup]
    truncated: bool = False
    truncation_message: Optional[str] = None


class FormulaFixerDiagnoseRequest(BaseModel):
    spreadsheet_id: str
    issue_id: str
    action: str = Field("dependency_path", description="dependency_path|diagnostic_tab|highlight_cells")


class FormulaFixerDiagnoseResponse(BaseModel):
    dependency_path: Optional[List[dict]] = None
    diagnostic_tab_data: Optional[dict] = None
    highlight_cells: Optional[List[str]] = None
    suspect_cells: Optional[List[dict]] = None


class FormulaFixerValidateItem(BaseModel):
    cell: str
    expected_original: Optional[str] = None
    proposed: str


class FormulaFixerValidateRequest(BaseModel):
    spreadsheet_id: str
    fixes: List[FormulaFixerValidateItem]


class FormulaFixerValidateResponse(BaseModel):
    valid: bool
    stale_cells: List[str] = Field(default_factory=list)


class FormulaFixerLogAppliedRequest(BaseModel):
    spreadsheet_id: str
    step_id: str
    fixes_applied: List[dict] = Field(default_factory=list)


class FormulaFixerLogAppliedResponse(BaseModel):
    logged: bool


def _ensure_enabled() -> None:
    if not saas_settings.GREMLIN_V2_ENABLED:
        raise HTTPException(status_code=status.HTTP_404_NOT_FOUND, detail="Gremlin v2 disabled")
    if not saas_settings.GREMLIN_V2_FORMULA_FIXER:
        raise HTTPException(status_code=status.HTTP_404_NOT_FOUND, detail="Formula fixer disabled")


@router.post("/scan", response_model=FormulaFixerScanResponse)
async def scan_formulas(
    request: FormulaFixerScanRequest,
    account_id: str = Depends(require_account),
):
    del account_id
    _ensure_enabled()

    raw_cells = [cell.model_dump() for cell in request.cells]
    cells, truncated, message = fixer_service.clamp_cells(raw_cells)

    existing_sheets = (
        request.sheet_titles
        or request.selected_sheets
        or sorted({cell.get("sheet") for cell in cells if cell.get("sheet")})
    )
    scan_id = str(uuid4())

    payload = fixer_service.build_scan_payload(cells, existing_sheets, scan_id)
    issue_groups = payload.get("issue_groups", [])

    redis = await get_redis()
    if redis:
        cache_key = fixer_service.SCAN_CACHE_KEY.format(scan_id=scan_id)
        await redis.setex(
            cache_key,
            fixer_service.SCAN_CACHE_TTL_SECONDS,
            fixer_service.build_scan_cache_payload(payload),
        )
        await redis.setex(
            fixer_service.LAST_SCAN_KEY.format(spreadsheet_id=request.spreadsheet_id),
            fixer_service.SCAN_CACHE_TTL_SECONDS,
            scan_id,
        )

    return FormulaFixerScanResponse(
        health=payload.get("health"),
        issue_groups=issue_groups,
        truncated=truncated,
        truncation_message=message,
    )


@router.post("/diagnose", response_model=FormulaFixerDiagnoseResponse)
async def diagnose_formula_issue(
    request: FormulaFixerDiagnoseRequest,
    account_id: str = Depends(require_account),
):
    del account_id
    _ensure_enabled()

    if ":" not in request.issue_id:
        raise HTTPException(status_code=status.HTTP_400_BAD_REQUEST, detail="Invalid issue_id")

    scan_id, issue_type = request.issue_id.split(":", 1)
    redis = await get_redis()
    if not redis:
        raise HTTPException(status_code=status.HTTP_503_SERVICE_UNAVAILABLE, detail="Diagnostics unavailable")

    raw = await redis.get(fixer_service.SCAN_CACHE_KEY.format(scan_id=scan_id))
    if not raw:
        raise HTTPException(status_code=status.HTTP_404_NOT_FOUND, detail="Scan data not found")

    cached = json.loads(raw)
    group = fixer_service.resolve_issue_group(cached, issue_type)
    if not group:
        raise HTTPException(status_code=status.HTTP_404_NOT_FOUND, detail="Issue group not found")

    cells_map = cached.get("cells") or {}
    cell_graph = cached.get("graph") or {}
    cells = group.get("cells") or []
    sample_cell = cells[0] if cells else None
    suspect_cells = (
        fixer_service.find_suspect_cells(sample_cell, cells_map, cell_graph)
        if sample_cell
        else []
    )

    if request.action == "highlight_cells":
        return FormulaFixerDiagnoseResponse(highlight_cells=cells)

    if request.action == "diagnostic_tab":
        trace = []
        if sample_cell:
            trace = gremlin_graph.trace_upstream(cell_graph, sample_cell, max_depth=10, cells_metadata=cells_map)
        rows = [[sample_cell or "", cells_map.get(sample_cell, {}).get("formula") or "", "root"]]
        for item in trace:
            rows.append([
                item.get("cell"),
                item.get("formula") or "",
                item.get("parent") or "",
            ])
        return FormulaFixerDiagnoseResponse(
            diagnostic_tab_data={
                "headers": ["Cell", "Formula", "Parent"],
                "rows": rows,
            },
            suspect_cells=suspect_cells or None,
        )

    trace = []
    if sample_cell:
        trace = gremlin_graph.trace_upstream(cell_graph, sample_cell, max_depth=10, cells_metadata=cells_map)
    dependency_path = []
    if sample_cell:
        dependency_path.append(
            {
                "cell": sample_cell,
                "formula": cells_map.get(sample_cell, {}).get("formula"),
                "references": cell_graph.get(sample_cell, {}).get("references", []),
            }
        )
    for item in trace:
        cell = item.get("cell")
        dependency_path.append(
            {
                "cell": cell,
                "formula": item.get("formula"),
                "references": cell_graph.get(cell, {}).get("references", []),
            }
        )

    return FormulaFixerDiagnoseResponse(
        dependency_path=dependency_path,
        suspect_cells=suspect_cells or None,
    )


@router.post("/validate", response_model=FormulaFixerValidateResponse)
async def validate_fixes(
    request: FormulaFixerValidateRequest,
    account_id: str = Depends(require_account),
):
    del account_id
    _ensure_enabled()

    redis = await get_redis()
    if not redis:
        return FormulaFixerValidateResponse(valid=True, stale_cells=[])

    scan_id_raw = await redis.get(
        fixer_service.LAST_SCAN_KEY.format(spreadsheet_id=request.spreadsheet_id)
    )
    if not scan_id_raw:
        return FormulaFixerValidateResponse(valid=True, stale_cells=[])

    scan_id = scan_id_raw.decode() if hasattr(scan_id_raw, "decode") else str(scan_id_raw)
    raw = await redis.get(fixer_service.SCAN_CACHE_KEY.format(scan_id=scan_id))
    if not raw:
        return FormulaFixerValidateResponse(valid=True, stale_cells=[])

    cached = json.loads(raw)
    cells_map = cached.get("cells") or {}

    stale: List[str] = []
    for fix in request.fixes:
        cell = fix.cell
        expected = fix.expected_original or ""
        if not expected:
            continue
        current = (cells_map.get(cell, {}) or {}).get("formula") or ""
        if current and expected != current:
            stale.append(cell)

    return FormulaFixerValidateResponse(valid=len(stale) == 0, stale_cells=stale)


@router.post("/log-applied", response_model=FormulaFixerLogAppliedResponse)
async def log_applied_fixes(
    request: FormulaFixerLogAppliedRequest,
    account_id: str = Depends(require_account),
):
    _ensure_enabled()
    logger.info(
        "Formula fixer applied: account=%s spreadsheet=%s step=%s fixes=%d",
        account_id,
        request.spreadsheet_id,
        request.step_id,
        len(request.fixes_applied or []),
    )
    return FormulaFixerLogAppliedResponse(logged=True)



import asyncio
import logging

log = logging.getLogger(__name__)


async def publish(topic: str, account_id: str, blocks: dict) -> None:
    """Thin wrapper to call API-layer publisher without circular imports."""
    try:
        # Lazy import to avoid circulars
        from ..api.v2.slack import publish as api_publish  # type: ignore

        await api_publish(topic, account_id, blocks, db=None)
    except Exception:
        log.exception("Slack publish failed topic=%s account=%s", topic, account_id)


def fire_and_forget(topic: str, account_id: str, blocks: dict) -> None:
    try:
        asyncio.create_task(publish(topic, account_id, blocks))
    except Exception:
        pass



"""Async worker for pushing contacts to HeyReach campaigns."""

from __future__ import annotations

import asyncio
import csv
import io
import json
import logging
import random
from dataclasses import dataclass
from datetime import datetime, timezone
from typing import Any, Dict, List, Optional

from g_gremlin.heyreach import (
    HeyReachClient,
    HeyReachProspect,
    HeyReachAuthError,
    HeyReachError,
    build_idempotency_key,
)

from ..api.progress import get_redis
from ..db import AsyncSessionLocal
from ..services.activation_utils import clean_string, normalize_status, get_row_number, should_skip_row
from ..services.email_notifications import send_heyreach_complete_notification
from ..services.heyreach_models import HeyReachConfig
from ..services.secret_storage import get_saas_secret_backend
from ..storage import get_storage_backend

logger = logging.getLogger(__name__)

META_KEY = "heyreach:job:{job_id}:meta"
STATE_KEY = "heyreach:job:{job_id}:state"
PROGRESS_KEY = "heyreach:job:{job_id}:progress"
SUMMARY_KEY = "heyreach:job:{job_id}:summary"
ERRORS_KEY = "heyreach:job:{job_id}:errors"
RESULTS_KEY = "heyreach:job:{job_id}:results"
CURSOR_KEY = "heyreach:job:{job_id}:cursor"

RESULT_TTL = 86400
MAX_ERROR_COUNT = 50

# Aliases for backward compatibility with internal references
_clean_string = clean_string
_normalize_status = normalize_status
_get_row_number = get_row_number
_should_skip_row = should_skip_row


@dataclass
class RowContext:
    row_number: int
    prospect: HeyReachProspect


class HeyReachWorker:
    """Process HeyReach push job asynchronously."""

    CHUNK_SIZE = 10

    def __init__(
        self,
        account_id: str,
        job_id: str,
        config: Dict[str, Any],
        rows: Optional[List[Dict[str, Any]]] = None,
        redis=None,
        sleep: Optional[callable] = None,
    ) -> None:
        self.account_id = account_id
        self.job_id = job_id
        self.config = HeyReachConfig.model_validate(config).model_dump()
        self.rows = rows or []
        self.redis = redis
        self.sleep = sleep or asyncio.sleep
        self.summary = {
            "total_rows": len(self.rows),
            "processed_rows": 0,
            "enrolled": 0,
            "failed": 0,
            "skipped": 0,
            "invalid": 0,
            "duplicate": 0,
        }
        self.errors: List[Dict[str, Any]] = []
        self.errors_truncated = False
        self._resume_cursor = 0

    async def _get_redis(self):
        if self.redis is not None:
            return self.redis
        return await get_redis()

    async def _set_state(self, state: str, message: Optional[str] = None) -> None:
        redis = await self._get_redis()
        if not redis:
            return
        await redis.set(STATE_KEY.format(job_id=self.job_id), state, ex=RESULT_TTL)
        await self._update_job_meta(status=state, message=message)

    async def _set_progress(self, progress: int) -> None:
        redis = await self._get_redis()
        if not redis:
            return
        await redis.set(PROGRESS_KEY.format(job_id=self.job_id), str(progress), ex=RESULT_TTL)
        await self._update_job_meta(progress=progress, processed_rows=self.summary["processed_rows"])

    async def _set_summary(self) -> None:
        redis = await self._get_redis()
        if not redis:
            return
        payload = json.dumps(self.summary)
        await redis.set(SUMMARY_KEY.format(job_id=self.job_id), payload, ex=RESULT_TTL)
        await redis.set(f"job:summary:{self.job_id}", payload, ex=RESULT_TTL)
        await self._update_job_meta(summary=self.summary)

    async def _set_errors(self) -> None:
        redis = await self._get_redis()
        if not redis:
            return
        payload = {
            "errors": self.errors,
            "errors_truncated": self.errors_truncated,
        }
        await redis.set(ERRORS_KEY.format(job_id=self.job_id), json.dumps(payload), ex=RESULT_TTL)
        await self._update_job_meta(errors=payload)

    async def _store_result(self, result: Dict[str, Any]) -> None:
        redis = await self._get_redis()
        if not redis:
            return
        await redis.rpush(RESULTS_KEY.format(job_id=self.job_id), json.dumps(result))
        await redis.expire(RESULTS_KEY.format(job_id=self.job_id), RESULT_TTL)

    async def _update_job_meta(self, **updates: Any) -> None:
        redis = await self._get_redis()
        if not redis:
            return
        key = f"job:{self.job_id}"
        raw = await redis.get(key)
        meta: Dict[str, Any]
        if raw:
            try:
                meta = json.loads(raw)
            except Exception:
                meta = {}
        else:
            meta = {}
        meta.update({k: v for k, v in updates.items() if v is not None})
        await redis.set(key, json.dumps(meta), ex=RESULT_TTL)

    def _record_error(self, row_number: int, status: str, error: str) -> None:
        if len(self.errors) >= MAX_ERROR_COUNT:
            self.errors_truncated = True
            return
        self.errors.append({
            "row_number": row_number,
            "status": status,
            "error": error,
        })

    def _increment_summary(self, status: str) -> None:
        if status in self.summary:
            self.summary[status] += 1
        self.summary["processed_rows"] = (
            self.summary["enrolled"]
            + self.summary["failed"]
            + self.summary["skipped"]
            + self.summary["invalid"]
            + self.summary["duplicate"]
        )

    async def _set_resume_cursor(self, cursor: int) -> None:
        redis = await self._get_redis()
        if not redis:
            return
        await redis.set(CURSOR_KEY.format(job_id=self.job_id), str(cursor), ex=RESULT_TTL)

    async def _load_rows(self) -> None:
        if self.rows:
            self.summary["total_rows"] = len(self.rows)
            return

        upload_id = self.config.get("upload_id")
        if not upload_id:
            return

        storage = get_storage_backend()
        if asyncio.iscoroutine(storage):
            storage = await storage
        content = await storage.read(upload_id)
        text = content.decode("utf-8-sig", errors="ignore")
        reader = csv.DictReader(io.StringIO(text))
        rows = list(reader)
        starting_row = int(self.config.get("starting_row") or 2)
        for idx, row in enumerate(rows):
            row["row_number"] = starting_row + idx
        self.rows = rows
        self.summary["total_rows"] = len(rows)

    async def _get_client(self) -> HeyReachClient:
        profile_id = self.config.get("profile_id")
        if not profile_id:
            raise HeyReachAuthError("Profile not configured")
        async with AsyncSessionLocal() as db:
            backend = get_saas_secret_backend()
            secret = await backend.get_secret_by_id(
                db=db,
                secret_id=profile_id,
                workspace_id=self.account_id,
                integration="heyreach",
            )
            workspace_secret = await backend.get_secret(
                db=db,
                integration="heyreach",
                key="workspace_id",
                workspace_id=self.account_id,
            )
        if not secret or not secret.value:
            raise HeyReachAuthError("HeyReach credentials not configured")
        workspace_id = None
        if workspace_secret and workspace_secret.value:
            workspace_id = _clean_string(workspace_secret.value)
        return HeyReachClient(api_key=str(secret.value), workspace_id=workspace_id)

    def _build_prospect(self, row: Dict[str, Any], mapping: Dict[str, Any]) -> Optional[HeyReachProspect]:
        def value_for(field: str) -> Optional[str]:
            column = mapping.get(field)
            if not column:
                return None
            value = row.get(column)
            text = _clean_string(value)
            return text if text else None

        email = value_for("email")
        linkedin_url = value_for("linkedin_url")
        if not email or not linkedin_url:
            return None

        custom_vars = {}
        for key, column in (mapping.get("custom_vars") or {}).items():
            if not column:
                continue
            raw = row.get(column)
            text = _clean_string(raw)
            if text:
                custom_vars[str(key)] = text

        return HeyReachProspect(
            email=email,
            linkedin_url=linkedin_url,
            first_name=value_for("first_name"),
            last_name=value_for("last_name"),
            company=value_for("company"),
            title=value_for("title"),
            sender_account_id=value_for("sender_account_id"),
            custom_vars=custom_vars,
        )

    def _build_batch_results(
        self,
        outcome,
        row_contexts: List[RowContext],
        campaign_id: str,
    ) -> List[Dict[str, Any]]:
        contexts_by_email: Dict[str, List[RowContext]] = {}
        for ctx in row_contexts:
            key = ctx.prospect.email.strip().lower()
            contexts_by_email.setdefault(key, []).append(ctx)

        def pop_ctx(email: str) -> Optional[RowContext]:
            key = (email or "").strip().lower()
            if not key:
                return None
            bucket = contexts_by_email.get(key)
            if not bucket:
                return None
            return bucket.pop(0)

        results: List[Dict[str, Any]] = []
        now_iso = datetime.now(timezone.utc).isoformat()

        for item in outcome.succeeded or []:
            ctx = pop_ctx(item.get("prospect_email", ""))
            if not ctx:
                continue
            status = _clean_string(item.get("status") or "enrolled")
            if status == "duplicate":
                self._increment_summary("duplicate")
            else:
                self._increment_summary("enrolled")
            results.append({
                "row_number": ctx.row_number,
                "status": status,
                "campaign_id": campaign_id,
                "provider_prospect_id": item.get("provider_prospect_id"),
                "pushed_at": now_iso,
            })

        for item in (outcome.failed or []) + (outcome.retryable_failed or []):
            ctx = pop_ctx(item.get("prospect_email", ""))
            if not ctx:
                continue
            error = _clean_string(item.get("error") or "Failed to enroll")
            self._record_error(ctx.row_number, "failed", error)
            self._increment_summary("failed")
            results.append({
                "row_number": ctx.row_number,
                "status": "failed",
                "campaign_id": campaign_id,
                "error": error,
                "pushed_at": now_iso,
            })

        for remaining in contexts_by_email.values():
            for ctx in remaining:
                error = "No enrollment result returned"
                self._record_error(ctx.row_number, "failed", error)
                self._increment_summary("failed")
                results.append({
                    "row_number": ctx.row_number,
                    "status": "failed",
                    "campaign_id": campaign_id,
                    "error": error,
                    "pushed_at": now_iso,
                })

        return results

    async def run(self) -> None:
        adapter: Optional[HeyReachClient] = None
        try:
            await self._set_state("processing", message="Starting HeyReach push")
            await self._load_rows()

            if not self.rows:
                await self._set_summary()
                await self._set_state("completed", message="No rows to process")
                await self._set_progress(100)
                return

            adapter = await self._get_client()
            campaign_id = self.config.get("campaign_id")
            mapping = self.config.get("mapping") or {}
            writeback = self.config.get("writeback") or {}
            starting_row = int(self.config.get("starting_row") or 2)
            total_rows = int(self.summary.get("total_rows") or len(self.rows))

            row_contexts: List[RowContext] = []
            seen_emails: set[str] = set()

            for idx, row in enumerate(self.rows):
                row_number = _get_row_number(row, idx, starting_row)

                if _should_skip_row(row, campaign_id, writeback):
                    await self._store_result({
                        "row_number": row_number,
                        "status": "skipped",
                        "campaign_id": campaign_id,
                        "error": "Already enrolled",
                    })
                    self._increment_summary("skipped")
                    await self._set_resume_cursor(idx + 1)
                    continue

                prospect = self._build_prospect(row, mapping)
                if not prospect:
                    await self._store_result({
                        "row_number": row_number,
                        "status": "invalid",
                        "campaign_id": campaign_id,
                        "error": "Missing required email or LinkedIn URL",
                    })
                    self._record_error(row_number, "invalid", "Missing required email or LinkedIn URL")
                    self._increment_summary("invalid")
                    await self._set_resume_cursor(idx + 1)
                    continue

                email_key = prospect.email.strip().lower()
                if email_key in seen_emails:
                    await self._store_result({
                        "row_number": row_number,
                        "status": "duplicate",
                        "campaign_id": campaign_id,
                        "error": "Duplicate email in selection",
                    })
                    self._increment_summary("duplicate")
                    await self._set_resume_cursor(idx + 1)
                    continue

                seen_emails.add(email_key)
                row_contexts.append(RowContext(row_number=row_number, prospect=prospect))

                if len(row_contexts) >= self.CHUNK_SIZE:
                    batch_prospects = [ctx.prospect for ctx in row_contexts]
                    idempotency_keys = [
                        build_idempotency_key(campaign_id, ctx.prospect.email)
                        for ctx in row_contexts
                    ]
                    outcome = adapter.enroll_prospects(
                        batch_prospects,
                        campaign_id,
                        idempotency_keys=idempotency_keys,
                    )
                    payloads = self._build_batch_results(outcome, row_contexts, campaign_id)
                    for payload in payloads:
                        await self._store_result(payload)
                    await self._set_summary()
                    await self._set_errors()
                    progress = int((self.summary["processed_rows"] / max(total_rows, 1)) * 100)
                    await self._set_progress(min(progress, 99))
                    row_contexts = []
                    await self._set_resume_cursor(idx + 1)
                    await self.sleep(0.5 * random.uniform(0.5, 1.5))

            if row_contexts:
                batch_prospects = [ctx.prospect for ctx in row_contexts]
                idempotency_keys = [
                    build_idempotency_key(campaign_id, ctx.prospect.email)
                    for ctx in row_contexts
                ]
                outcome = adapter.enroll_prospects(
                    batch_prospects,
                    campaign_id,
                    idempotency_keys=idempotency_keys,
                )
                payloads = self._build_batch_results(outcome, row_contexts, campaign_id)
                for payload in payloads:
                    await self._store_result(payload)

            await self._set_summary()
            await self._set_errors()
            await self._set_progress(100)
            await self._set_state("completed", message="HeyReach push completed")
            await self._send_completion_notification()
        except (HeyReachAuthError, HeyReachError) as exc:
            await self._set_state("failed", message=str(exc))
            await self._set_progress(100)
            raise
        except Exception as exc:
            await self._set_state("failed", message=str(exc))
            await self._set_progress(100)
            raise

    async def _send_completion_notification(self) -> None:
        """Send email notification when job completes."""
        notify_email = self.config.get("notify_email")
        if not notify_email:
            return
        spreadsheet_url = self.config.get("spreadsheet_url")
        try:
            await send_heyreach_complete_notification(
                user_email=notify_email,
                job_name="HeyReach LinkedIn Push",
                job_id=self.job_id,
                total_rows=self.summary.get("total_rows", 0),
                enrolled_count=self.summary.get("enrolled", 0),
                failed_count=self.summary.get("failed", 0),
                skipped_count=self.summary.get("skipped", 0),
                duplicate_count=self.summary.get("duplicate", 0),
                invalid_count=self.summary.get("invalid", 0),
                spreadsheet_url=spreadsheet_url,
            )
        except Exception as exc:
            logger.warning(f"Failed to send completion notification: {exc}")



"""
Write enrichment results back to Salesforce.

Features:
- FLS validation
- Bulk API for performance
- Audit trail via FM_DataSource__c
- Dry-run mode
- Detailed error reporting
- Archival of enriched data for audit trail and recovery
"""

import logging
from dataclasses import dataclass, field
from datetime import datetime
from typing import Any, Dict, List, Optional, Set

import sqlalchemy as sa
from sqlalchemy.dialects.postgresql import insert
from sqlalchemy.exc import IntegrityError
from sqlalchemy.ext.asyncio import AsyncSession

from ...services.salesforce_gateway import SalesforceGateway
from .contact_enrichment_coordinator import EnrichmentMatch
from ...model_defs.enrichment_vault import EnrichmentVault
from ...models import EnrichmentUsage

logger = logging.getLogger(__name__)


@dataclass
class EnrichmentWriteResult:
    """Result of writing enrichments to Salesforce."""

    # Overall status
    success: bool
    total_records: int
    records_written: int
    records_failed: int

    # Detailed results
    successful_ids: List[str] = field(default_factory=list)
    failed_ids: List[str] = field(default_factory=list)
    errors: List[Dict[str, str]] = field(
        default_factory=list
    )  # [{"id": "...", "error": "..."}]

    # Statistics
    fields_updated: Dict[str, int] = field(
        default_factory=dict
    )  # {"email": 50, "phone": 30, ...}

    # Metadata
    dry_run: bool = False
    audit_trail: Optional[str] = None


class EnrichmentWriter:
    """
    Write enrichment results to Salesforce.

    Responsibilities:
    - Validate FLS permissions
    - Build update records
    - Use Bulk API for performance
    - Add audit trail metadata
    - Handle partial failures gracefully
    """

    # Custom field for audit trail (will be created if doesn't exist)
    AUDIT_FIELD_NAME = "FM_DataSource__c"

    def __init__(
        self,
        gateway: SalesforceGateway,
        scanner=None,
        validate_fls: bool = True,
        add_audit_trail: bool = True,
    ):
        """
        Initialize writer.

        Args:
            gateway: Salesforce gateway for writes
            scanner: MissingFieldScanner instance for field mapping (optional)
            validate_fls: Check FLS before writing
            add_audit_trail: Add FM_DataSource__c field to track enrichment source
        """
        self.gateway = gateway
        self.scanner = scanner
        self.validate_fls = validate_fls
        self.add_audit_trail = add_audit_trail

    async def write_enrichments(
        self,
        account_id: str,
        matches: List[EnrichmentMatch],
        dry_run: bool = False,
        db: Optional[AsyncSession] = None,
        usage_id: Optional[int] = None,
    ) -> EnrichmentWriteResult:
        """
        Write enrichment matches to Salesforce.

        Args:
            account_id: FoundryMatch account ID
            matches: List of enrichment matches to write
            dry_run: If True, validate but don't actually write
            db: Optional database session for archiving enriched data
            usage_id: Optional EnrichmentUsage ID for archiving

        Returns:
            EnrichmentWriteResult with detailed status
        """
        result = EnrichmentWriteResult(
            success=False,
            total_records=len(matches),
            records_written=0,
            records_failed=0,
            dry_run=dry_run,
        )

        if not matches:
            logger.info("No matches to write")
            result.success = True
            return result

        logger.info(
            f"Writing {len(matches)} enrichments to Salesforce (dry_run={dry_run})"
        )

        # Check if archival is enabled (only if db session provided)
        archive_enabled = False
        if db and usage_id and not dry_run:
            archive_enabled = await self._should_archive(db, account_id)
            if archive_enabled:
                logger.info(f"Enrichment archival enabled for account {account_id}")

        # Group by object type (Contact vs Lead)
        by_object = self._group_by_object(matches)

        # Process each object type
        for sfdc_object, object_matches in by_object.items():
            logger.info(f"Processing {len(object_matches)} {sfdc_object} records")

            # Validate FLS if enabled
            if self.validate_fls:
                fls_result = await self._validate_fls(
                    account_id, sfdc_object, object_matches
                )
                if not fls_result["valid"]:
                    # Remove fields without update permission
                    object_matches = self._filter_by_fls(
                        object_matches, fls_result["updateable_fields"]
                    )
                    if not object_matches:
                        logger.error(f"No updateable fields for {sfdc_object}")
                        continue

            # Build update records
            update_records = self._build_update_records(object_matches)

            if not update_records:
                logger.warning(f"No records to update for {sfdc_object}")
                continue

            # Fetch before values if archival is enabled
            before_lookup: Dict[str, Dict[str, Any]] = {}
            if archive_enabled:
                record_ids = [r["Id"] for r in update_records]
                field_names = list(
                    set(k for r in update_records for k in r.keys() if k != "Id")
                )
                before_lookup = await self._fetch_current_values(
                    sfdc_object, record_ids, field_names
                )
                logger.debug(f"Fetched before values for {len(before_lookup)} records")

            # Write to Salesforce (unless dry run)
            if dry_run:
                logger.info(
                    f"DRY RUN: Would update {len(update_records)} {sfdc_object} records"
                )
                result.records_written += len(update_records)
                result.successful_ids.extend([r["Id"] for r in update_records])
            else:
                write_result = await self._write_bulk(sfdc_object, update_records)
                result.records_written += write_result["success_count"]
                result.records_failed += write_result["failure_count"]
                result.successful_ids.extend(write_result["successful_ids"])
                result.failed_ids.extend(write_result["failed_ids"])
                result.errors.extend(write_result["errors"])

                # Archive successful writes if enabled
                if archive_enabled and write_result["success_count"] > 0:
                    try:
                        await self._archive_to_vault(
                            db,
                            usage_id=usage_id,
                            account_id=account_id,
                            sfdc_object=sfdc_object,
                            update_records=update_records,
                            write_result=write_result,
                            before_lookup=before_lookup,
                        )
                    except Exception as e:
                        logger.error(f"Archive failed but continuing: {e}")
                        # Don't fail the whole write if archival fails

            # Track field statistics
            for match in object_matches:
                for field_name in match.fields_to_update:
                    result.fields_updated[field_name] = (
                        result.fields_updated.get(field_name, 0) + 1
                    )

        # Set overall success
        result.success = result.records_failed == 0

        # Generate audit trail
        if self.add_audit_trail:
            result.audit_trail = self._generate_audit_trail()

        logger.info(
            f"Write complete: {result.records_written} succeeded, {result.records_failed} failed "
            f"(dry_run={dry_run})"
        )

        return result

    async def _validate_fls(
        self,
        account_id: str,
        sfdc_object: str,
        matches: List[EnrichmentMatch],
    ) -> Dict:
        """
        Validate field-level security for all fields we want to update.

        Returns:
            {"valid": bool, "updateable_fields": Set[str], "errors": List[str]}
        """
        # Collect all fields we want to update
        fields_to_check = set()
        for match in matches:
            for logical_field in match.fields_to_update:
                # Map logical field to SFDC field name
                sfdc_field = self._logical_to_sfdc_field(logical_field, sfdc_object)
                fields_to_check.add(sfdc_field)

        if self.add_audit_trail:
            fields_to_check.add(self.AUDIT_FIELD_NAME)

        logger.debug(f"Validating FLS for fields: {fields_to_check}")

        # Get field permissions via describe
        try:
            describe_result = await self.gateway.describe(sfdc_object)

            # Build field map: {field_name.lower(): field_metadata}
            field_map = {
                f["name"].lower(): f
                for f in describe_result.get("fields", [])
                if f.get("name")
            }

            # Check updateable flag for each field
            updateable_fields = set()
            errors = []

            for sfdc_field in fields_to_check:
                field_lower = sfdc_field.lower()
                field_meta = field_map.get(field_lower)

                if not field_meta:
                    errors.append(f"Field {sfdc_field} not found on {sfdc_object}")
                    logger.warning(f"Field {sfdc_field} not found on {sfdc_object}")
                    continue

                if not field_meta.get("updateable", False):
                    errors.append(
                        f"Field {sfdc_field} is not updateable (FLS permission denied)"
                    )
                    logger.warning(
                        f"Field {sfdc_field} on {sfdc_object} is not updateable"
                    )
                    continue

                # Field is updateable
                updateable_fields.add(sfdc_field)

            # Return validation result
            if not updateable_fields and fields_to_check:
                # All fields blocked by FLS
                return {
                    "valid": False,
                    "updateable_fields": updateable_fields,
                    "errors": errors,
                }

            return {
                "valid": True,
                "updateable_fields": updateable_fields,
                "errors": errors,  # Include warnings about non-updateable fields
            }

        except Exception as e:
            logger.error(f"FLS validation failed: {e}")
            return {
                "valid": False,
                "updateable_fields": set(),
                "errors": [f"Describe API failed: {str(e)}"],
            }

    def _filter_by_fls(
        self,
        matches: List[EnrichmentMatch],
        updateable_fields: Set[str],
    ) -> List[EnrichmentMatch]:
        """
        Filter matches to only include fields that are updateable.

        Returns new list of matches with filtered fields_to_update.
        """
        filtered = []

        for match in matches:
            # Filter fields_to_update
            allowed_fields = set()
            for logical_field in match.fields_to_update:
                sfdc_field = self._logical_to_sfdc_field(
                    logical_field, match.candidate.sfdc_object
                )
                if sfdc_field in updateable_fields:
                    allowed_fields.add(logical_field)

            if allowed_fields:
                # Create new match with filtered fields
                import copy

                new_match = copy.deepcopy(match)
                new_match.fields_to_update = allowed_fields
                filtered.append(new_match)

        return filtered

    def _build_update_records(self, matches: List[EnrichmentMatch]) -> List[Dict]:
        """
        Build Salesforce update records from matches.

        Returns:
            List of dicts ready for Bulk API: [{"Id": "...", "Email": "...", ...}]
        """
        records = []
        active_enrichable_fields: Set[str] = set()
        if self.scanner and hasattr(self.scanner, "get_enrichable_fields"):
            try:
                active_enrichable_fields = set(self.scanner.get_enrichable_fields())
            except Exception:  # pragma: no cover - defensive guard
                logger.debug(
                    "Failed to read enrichable fields from scanner", exc_info=True
                )

        for match in matches:
            record = {"Id": match.candidate.sfdc_id}

            # Add enriched fields
            sfdc_object = match.candidate.sfdc_object

            if "email" in match.fields_to_update and match.enriched_email:
                field_name = self._logical_to_sfdc_field("email", sfdc_object)
                record[field_name] = match.enriched_email

            if "phone" in match.fields_to_update and match.enriched_phone:
                field_name = self._logical_to_sfdc_field("phone", sfdc_object)
                record[field_name] = match.enriched_phone

            if "mobile_phone" in match.fields_to_update and match.enriched_mobile_phone:
                field_name = self._logical_to_sfdc_field("mobile_phone", sfdc_object)
                record[field_name] = match.enriched_mobile_phone

            if "title" in match.fields_to_update and match.enriched_title:
                field_name = self._logical_to_sfdc_field("title", sfdc_object)
                record[field_name] = match.enriched_title

            if (
                "linkedin_url" in match.fields_to_update
                and match.enriched_linkedin_url
                and "linkedin_url" in active_enrichable_fields
            ):
                field_name = self._logical_to_sfdc_field("linkedin_url", sfdc_object)
                if field_name.lower() == "linkedin_url":
                    logger.debug(
                        "Skipping linkedin_url update due to unresolved mapping (sfdc_object=%s)",
                        sfdc_object,
                    )
                else:
                    record[field_name] = match.enriched_linkedin_url

            # Add audit trail
            if self.add_audit_trail:
                record[self.AUDIT_FIELD_NAME] = self._generate_audit_trail()

            # Only include record if we're actually updating something
            if len(record) > 1:  # More than just Id
                records.append(record)

        return records

    async def _write_bulk(
        self,
        sfdc_object: str,
        records: List[Dict],
    ) -> Dict:
        """
        Write records via Salesforce Bulk API.

        Returns:
            {
                "success_count": int,
                "failure_count": int,
                "successful_ids": List[str],
                "failed_ids": List[str],
                "errors": List[Dict],
            }
        """
        logger.info(f"Writing {len(records)} {sfdc_object} records via Bulk API")

        try:
            # Use gateway's bulk update method
            result = await self.gateway.bulk_update(sfdc_object, records)

            if result.get("error"):
                logger.error(f"Bulk update failed: {result['error']}")
                return {
                    "success_count": 0,
                    "failure_count": len(records),
                    "successful_ids": [],
                    "failed_ids": [r["Id"] for r in records],
                    "errors": [
                        {"id": r["Id"], "error": result["error"]} for r in records
                    ],
                }

            # Process results
            successes = result.get("successfulResults", [])
            failures = result.get("failedResults", [])

            return {
                "success_count": len(successes),
                "failure_count": len(failures),
                "successful_ids": [s.get("id") for s in successes if s.get("id")],
                "failed_ids": [f.get("id") for f in failures if f.get("id")],
                "errors": [
                    {
                        "id": f.get("id"),
                        "error": f.get("errors", [{}])[0].get(
                            "message", "Unknown error"
                        ),
                    }
                    for f in failures
                ],
            }

        except Exception as e:
            logger.error(f"Bulk write exception: {e}")
            return {
                "success_count": 0,
                "failure_count": len(records),
                "successful_ids": [],
                "failed_ids": [r["Id"] for r in records],
                "errors": [{"id": r["Id"], "error": str(e)} for r in records],
            }

    @staticmethod
    def _group_by_object(
        matches: List[EnrichmentMatch],
    ) -> Dict[str, List[EnrichmentMatch]]:
        """Group matches by Salesforce object type (Contact vs Lead)."""
        by_object = {}

        for match in matches:
            sfdc_object = match.candidate.sfdc_object
            if sfdc_object not in by_object:
                by_object[sfdc_object] = []
            by_object[sfdc_object].append(match)

        return by_object

    def _logical_to_sfdc_field(self, logical_field: str, sfdc_object: str) -> str:
        """
        Map logical field name to actual Salesforce field name.

        Uses scanner's field mapping if available (supports custom fields),
        otherwise falls back to standard mapping.
        """
        # Delegate to scanner if available
        if self.scanner:
            return self.scanner._get_field_name(sfdc_object, logical_field)

        # Fallback to standard mapping
        mapping = {
            "email": "Email",
            "phone": "Phone",
            "mobile_phone": "MobilePhone",
            "title": "Title",
        }
        return mapping.get(logical_field, logical_field)

    @staticmethod
    def _generate_audit_trail() -> str:
        """Generate audit trail string for FM_DataSource__c field."""
        today = datetime.now().strftime("%Y-%m-%d")
        return f"PDL_{today}"

    # -------------------------------------------------------------------------
    # Archival methods (for audit trail and data recovery)
    # -------------------------------------------------------------------------

    @staticmethod
    async def _should_archive(db: AsyncSession, account_id: str) -> bool:
        """Check if account has archival enabled."""
        try:
            row = await db.execute(
                sa.text(
                    "SELECT enrichment_archival_enabled FROM accounts WHERE id = :aid"
                ),
                {"aid": account_id},
            )
            val = row.scalar_one_or_none()
            return bool(val) if val is not None else True
        except Exception as e:
            logger.warning(
                f"Failed to check archival setting for account {account_id}: {e}"
            )
            return True  # Default to enabled if check fails

    async def _fetch_current_values(
        self,
        sfdc_object: str,
        record_ids: List[str],
        field_names: List[str],
    ) -> Dict[str, Dict[str, Any]]:
        """Fetch current field values from SFDC before overwriting (for before/after audit)."""
        if not record_ids or not field_names:
            return {}

        try:
            # Build SOQL: SELECT Id, Email, Title, MobilePhone FROM Lead WHERE Id IN (...)
            fields_str = ", ".join(["Id"] + sorted(set(field_names)))
            ids_str = "','".join(record_ids)
            soql = f"SELECT {fields_str} FROM {sfdc_object} WHERE Id IN ('{ids_str}')"

            records = await self.gateway.query(soql)
            result = {}
            for rec in records.get("records", []):
                rec_id = rec.get("Id")
                if rec_id:
                    result[rec_id] = {k: v for k, v in rec.items() if k != "Id"}
            return result
        except Exception as e:
            logger.warning(f"Failed to fetch current values for archival: {e}")
            return {}

    @staticmethod
    async def _archive_to_vault(
        db: AsyncSession,
        *,
        usage_id: int,
        account_id: str,
        sfdc_object: str,
        update_records: List[Dict],
        write_result: Dict,
        before_lookup: Dict[str, Dict[str, Any]],
    ) -> None:
        """
        Field-level archive. Writes one row per (record, field).

        Args:
            db: Database session
            usage_id: EnrichmentUsage.id (ledger reference)
            account_id: Account ID
            sfdc_object: SFDC object type (Lead, Contact, Account)
            update_records: Records that were sent to SFDC [{"Id":"003...", "Email":"x", ...}]
            write_result: Result from _write_bulk with success_ids
            before_lookup: Dict of {record_id: {field: value}} for before state
        """
        success_ids = set(write_result.get("successful_ids", []))

        rows = []
        for rec in update_records:
            rid = rec.get("Id") or rec.get("id")
            if not rid:
                continue
            if success_ids and rid not in success_ids:
                continue  # Only archive successful writes

            before_map = before_lookup.get(rid, {})
            for k, v in rec.items():
                if k in ("Id", "id"):
                    continue
                vb = before_map.get(k)  # May be None if no snapshot was taken
                rows.append(
                    {
                        "usage_id": usage_id,
                        "account_id": account_id,
                        "sfdc_object": sfdc_object,
                        "sfdc_record_id": rid,
                        "field_name": k,
                        "value_before": None if vb is None else str(vb),
                        "value_after": "" if v is None else str(v),
                        "provider_confidence": None,  # Can be filled if you carry scores forward
                    }
                )

        if not rows:
            return

        # Use INSERT ... ON CONFLICT DO NOTHING for idempotency
        stmt = insert(EnrichmentVault).values(rows)
        stmt = stmt.on_conflict_do_nothing(
            index_elements=["usage_id", "sfdc_record_id", "field_name"]
        )

        try:
            await db.execute(stmt)
            await db.execute(
                sa.update(EnrichmentUsage)
                .where(EnrichmentUsage.id == usage_id)
                .values(has_archived_data=True)
            )
            logger.info(
                f"Archived {len(rows)} field-level records to vault for usage_id={usage_id}"
            )
        except IntegrityError as e:
            # Safe to ignore duplicate rows caused by retries
            logger.debug(f"Vault archive conflict (expected on retry): {e}")
        except Exception as e:
            logger.error(f"Failed to archive to vault: {e}")



import pandas as pd

from fmatch.saas.intelligence.b2b_detectors import (
    B2BFieldDetector,
    auto_map_b2b,
    fingerprint_match,
)
from tests.fixtures.b2b_test_data import (
    SALESFORCE_ACCOUNTS,
    MESSY_CONTACTS
)


def test_detects_obvious_fields():
    detector = B2BFieldDetector(SALESFORCE_ACCOUNTS)
    results = detector.detect_all()

    assert results["account_name"].column == "Account Name"
    assert results["account_name"].confidence >= 0.8
    assert results["domain"].column == "Website"
    assert results["domain"].confidence >= 0.8


def test_handles_messy_headers():
    detector = B2BFieldDetector(MESSY_CONTACTS)
    results = detector.detect_all()

    assert results["account_name"].column == "Organization"
    assert results["account_name"].confidence >= 0.6
    assert results["account_name"].column != "Company_OLD_DO_NOT_USE"


def test_cross_validation_boost():
    df = pd.DataFrame(
        {
            "Company": ["Acme Corp", "Beta LLC", "Gamma Ltd"],
            "Website": ["acme.com", "beta.com", "gamma.com"],
            "Email": ["sales@acme.com", "info@beta.com", "hello@gamma.com"],
        }
    )
    detector = B2BFieldDetector(df)
    results = detector.detect_all()

    assert "cross_validation" in results["domain"].explanation


def test_fingerprint_matching():
    headers = list(SALESFORCE_ACCOUNTS.columns)
    match = fingerprint_match(headers)

    assert match is not None
    assert match[0] == "salesforce_accounts"


def test_requires_confirmation_trigger():
    source_df = pd.DataFrame({"Company": ["Foo Co", "Bar Co", "Baz Co"]})
    reference_df = pd.DataFrame({"Organization": ["Foo Co", "Bar Co", "Baz Co"]})

    result = auto_map_b2b(source_df, reference_df)

    assert result["requires_confirmation"] is True
    assert result["method"] == "detection"


"""Trusted dev-mode gating helpers.

These guards prevent single-env-var bypasses in public wheel installs.
"""

from __future__ import annotations

import os
from pathlib import Path

_TRUTHY = {"1", "true", "yes", "on"}


def _env_true(name: str) -> bool:
    return os.getenv(name, "").strip().lower() in _TRUTHY


def is_trusted_dev_context(module_file: Path | None = None) -> bool:
    """
    Return True only when running from a source checkout.

    Trusted context is defined as a parent directory that contains both
    `.git` and `pyproject.toml`. This does not exist in normal wheel installs.
    """
    target = (module_file or Path(__file__)).resolve()
    for parent in target.parents:
        if (parent / ".git").exists() and (parent / "pyproject.toml").exists():
            return True
        # Stop once we reach typical installed package roots.
        if parent.name in {"site-packages", "dist-packages"}:
            break
    return False


def is_dev_mode_enabled(module_file: Path | None = None) -> bool:
    """
    Dev mode requires two signals:
    1) `G_GREMLIN_DEV_MODE=1`
    2) trusted local source-checkout context
    """
    return _env_true("G_GREMLIN_DEV_MODE") and is_trusted_dev_context(module_file)




"""
Salesforce metadata smoke assertions.
"""

from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Dict, List

from g_gremlin import sfdc as sfdc_helpers
from g_gremlin import sfdc_schema as sfdc_schema_helpers


@dataclass
class SmokeAssertion:
    type: str
    params: Dict[str, Any]


@dataclass
class SmokeResult:
    assertion: SmokeAssertion
    passed: bool
    detail: str = ""

    @property
    def skipped(self) -> bool:
        return self.detail.startswith("skipped:")

    def to_dict(self) -> Dict[str, Any]:
        return {
            "type": self.assertion.type,
            "params": self.assertion.params,
            "passed": self.passed,
            "detail": self.detail,
        }


@dataclass
class SmokeReport:
    results: List[SmokeResult]

    @property
    def all_passed(self) -> bool:
        return all(r.passed or r.skipped for r in self.results)

    @property
    def summary(self) -> Dict[str, int]:
        total = len(self.results)
        passed = len([r for r in self.results if r.passed])
        failed = len([r for r in self.results if (not r.passed) and (not r.skipped)])
        skipped = len([r for r in self.results if r.skipped])
        return {"total": total, "passed": passed, "failed": failed, "skipped": skipped}

    def to_dict(self) -> Dict[str, Any]:
        return {"summary": self.summary, "results": [r.to_dict() for r in self.results]}


def load_smoke_assertions(pack_yaml_dict: Dict[str, Any]) -> List[SmokeAssertion]:
    items = pack_yaml_dict.get("smoke") or []
    assertions: List[SmokeAssertion] = []
    if isinstance(items, list):
        for item in items:
            if not isinstance(item, dict):
                continue
            atype = item.get("type") or item.get("assertion")
            params = item.get("params") or {k: v for k, v in item.items() if k not in {"type", "assertion"}}
            if atype:
                assertions.append(SmokeAssertion(type=str(atype), params=params))
    return assertions


def run_smoke(assertions: List[SmokeAssertion], *, org_alias: str | None = None) -> SmokeReport:
    results: List[SmokeResult] = []
    rule_cache: List[Dict[str, Any]] | None = None

    for assertion in assertions:
        atype = assertion.type
        params = assertion.params or {}

        if atype == "field_exists":
            results.append(_assert_field_exists(params, org_alias=org_alias))
            continue

        if atype == "permset_exists":
            results.append(_assert_permset_exists(params, org_alias=org_alias))
            continue

        if atype == "queue_exists":
            results.append(_assert_queue_exists(params, org_alias=org_alias))
            continue

        if atype == "validation_rule_exists":
            rule_cache = rule_cache or _fetch_validation_rules(org_alias=org_alias)
            results.append(_assert_validation_rule_exists(params, rule_cache))
            continue

        if atype == "record_type_exists":
            results.append(_assert_record_type_exists(params, org_alias=org_alias))
            continue

        raise ValueError(f"Unknown smoke assertion type: {atype}")

    return SmokeReport(results=results)


# ---- assertion helpers ----


def _assert_field_exists(params: Dict[str, Any], *, org_alias: str | None) -> SmokeResult:
    target = str(params.get("target") or params.get("field") or "")
    if not target or "." not in target:
        return SmokeResult(
            assertion=SmokeAssertion(type="field_exists", params=params),
            passed=False,
            detail="missing target (expected Object.Field__c)",
        )
    obj, field = target.split(".", 1)
    try:
        identity = sfdc_helpers.get_org_identity(org_alias=org_alias)
        schema_info = sfdc_schema_helpers.get_schema(obj, identity=identity, org_alias=org_alias)
    except Exception as exc:
        return SmokeResult(
            assertion=SmokeAssertion(type="field_exists", params=params),
            passed=False,
            detail=str(exc),
        )
    passed = field in schema_info.fields
    detail = "" if passed else f"field not found: {target}"
    return SmokeResult(assertion=SmokeAssertion(type="field_exists", params=params), passed=passed, detail=detail)


def _assert_permset_exists(params: Dict[str, Any], *, org_alias: str | None) -> SmokeResult:
    name = str(params.get("target") or params.get("name") or "")
    if not name:
        return SmokeResult(
            assertion=SmokeAssertion(type="permset_exists", params=params),
            passed=False,
            detail="missing permset name",
        )
    soql = (
        "SELECT Id FROM PermissionSet "
        f"WHERE Name = '{sfdc_helpers.escape_soql_value(name)}' LIMIT 1"
    )
    try:
        rows = sfdc_helpers.run_soql(soql, org_alias=org_alias)
    except Exception as exc:
        return SmokeResult(
            assertion=SmokeAssertion(type="permset_exists", params=params),
            passed=False,
            detail=str(exc),
        )
    passed = bool(rows)
    detail = "" if passed else f"permission set not found: {name}"
    return SmokeResult(assertion=SmokeAssertion(type="permset_exists", params=params), passed=passed, detail=detail)


def _assert_queue_exists(params: Dict[str, Any], *, org_alias: str | None) -> SmokeResult:
    name = str(params.get("target") or params.get("name") or "")
    if not name:
        return SmokeResult(
            assertion=SmokeAssertion(type="queue_exists", params=params),
            passed=False,
            detail="missing queue name",
        )
    escaped = sfdc_helpers.escape_soql_value(name)
    soql = (
        "SELECT Id FROM Group WHERE Type = 'Queue' AND "
        f"(DeveloperName = '{escaped}' OR Name = '{escaped}') LIMIT 1"
    )
    try:
        rows = sfdc_helpers.run_soql(soql, org_alias=org_alias)
    except Exception as exc:
        return SmokeResult(
            assertion=SmokeAssertion(type="queue_exists", params=params),
            passed=False,
            detail=str(exc),
        )
    passed = bool(rows)
    detail = "" if passed else f"queue not found: {name}"
    return SmokeResult(assertion=SmokeAssertion(type="queue_exists", params=params), passed=passed, detail=detail)


def _fetch_validation_rules(org_alias: str | None) -> List[Dict[str, Any]]:
    try:
        return list(sfdc_helpers.org_list_metadata("ValidationRule", org_alias=org_alias) or [])
    except Exception:
        return []


def _assert_validation_rule_exists(params: Dict[str, Any], rules: List[Dict[str, Any]]) -> SmokeResult:
    target = str(params.get("target") or params.get("name") or "")
    if not target or "." not in target:
        return SmokeResult(
            assertion=SmokeAssertion(type="validation_rule_exists", params=params),
            passed=False,
            detail="missing target (expected Object.Rule)",
        )
    target_lower = target.lower()
    for item in rules:
        full_name = str(item.get("fullName") or item.get("full_name") or item.get("name") or "")
        if full_name.lower() == target_lower:
            return SmokeResult(
                assertion=SmokeAssertion(type="validation_rule_exists", params=params),
                passed=True,
                detail="",
            )
    return SmokeResult(
        assertion=SmokeAssertion(type="validation_rule_exists", params=params),
        passed=False,
        detail=f"validation rule not found: {target}",
    )


def _assert_record_type_exists(params: Dict[str, Any], *, org_alias: str | None) -> SmokeResult:
    target = str(params.get("target") or params.get("name") or "")
    if not target or "." not in target:
        return SmokeResult(
            assertion=SmokeAssertion(type="record_type_exists", params=params),
            passed=False,
            detail="missing target (expected Object.RecordType)",
        )
    obj, record_type = target.split(".", 1)
    escaped_obj = sfdc_helpers.escape_soql_value(obj)
    escaped_rt = sfdc_helpers.escape_soql_value(record_type)
    soql = (
        "SELECT Id FROM RecordType "
        f"WHERE SObjectType = '{escaped_obj}' "
        f"AND (DeveloperName = '{escaped_rt}' OR Name = '{escaped_rt}') LIMIT 1"
    )
    try:
        rows = sfdc_helpers.run_soql(soql, org_alias=org_alias)
    except Exception as exc:
        return SmokeResult(
            assertion=SmokeAssertion(type="record_type_exists", params=params),
            passed=False,
            detail=str(exc),
        )
    passed = bool(rows)
    detail = "" if passed else f"record type not found: {target}"
    return SmokeResult(
        assertion=SmokeAssertion(type="record_type_exists", params=params),
        passed=passed,
        detail=detail,
    )



from __future__ import annotations

from typing import Any, Dict
import re


_EXT_RX = re.compile(r"(?:ext\.?|extension|x|;ext=|#)\s*\d{1,6}\b|,\d{1,6}$", re.I)


def _strip_extension(num: str) -> str:
    if not num:
        return ""
    return _EXT_RX.sub("", num).strip()


def _e164_fallback(num: str, country: str) -> str:
    core = _strip_extension(num)
    digits = "".join(ch for ch in (core or "") if ch.isdigit())
    if not digits:
        return ""
    if len(digits) < 7 or len(digits) > 15:
        return ""
    country_upper = (country or "US").upper()
    if country_upper in {"US", "CA"}:
        if len(digits) == 11 and digits.startswith("1"):
            return f"+{digits}"
        if len(digits) == 10:
            return f"+1{digits}"
    return f"+{digits}"


def fix_phone_format(text: str, *, country: str = "US") -> Dict[str, Any]:
    try:
        import phonenumbers  # optional dep

        p = phonenumbers.parse(text or "", country or "US")
        if not phonenumbers.is_valid_number(p):
            return {"value": ""}
        return {
            "value": phonenumbers.format_number(
                p, phonenumbers.PhoneNumberFormat.E164
            )
        }
    except Exception:
        return {"value": _e164_fallback(text, country or "US")}



"""Helper utilities to enrich promotion candidates before scoring."""

from __future__ import annotations

import logging
import threading
import time
from typing import Any, Dict, Optional, TYPE_CHECKING, Tuple

from sqlalchemy import text
from sqlalchemy.orm import Session

from ...utils_normalize import to_domain

if TYPE_CHECKING:  # pragma: no cover
    from .promotion_service import PromotionCandidateInput

log = logging.getLogger(__name__)

_TABLE_CACHE: Dict[str, bool] = {}
_LOOKUP_CACHE_TTL = 86_400  # 24 hours
_LOOKUP_CACHE: Dict[Tuple[str, Tuple[Tuple[str, Any], ...]], Tuple[float, Any]] = {}
_CACHE_LOCK = threading.Lock()
_LOOKUP_SEMAPHORE = threading.BoundedSemaphore(8)


def hydrate_candidate_facts(
    session: Session,
    *,
    candidate: "PromotionCandidateInput",
    target_snapshot: Optional[Dict[str, Any]],
) -> Dict[str, Any]:
    """
    Fill in missing authority and website facts for a promotion candidate.

    Hydration uses a per-process cache with a 24 hour TTL to avoid repeated lookups when
    large batches run, and records metadata so the caller can track hydration misses.
    """
    payload: Dict[str, Any] = dict(candidate.engine_payload or {})
    nested = (
        payload.get("candidate") if isinstance(payload.get("candidate"), dict) else {}
    )
    hydrate_meta: Dict[str, Any] = {
        "site_source": None,
        "country_source": None,
        "miss_reasons": [],
        "sec_match_type": None,
    }

    target_domain = None
    target_ticker = None
    target_country = None
    if target_snapshot:
        target_domain = target_snapshot.get("domain_claimed")
        target_ticker = target_snapshot.get("target_ticker") or target_snapshot.get(
            "ticker"
        )
        target_country = target_snapshot.get("target_country") or target_snapshot.get(
            "country_code"
        )

    site = _extract_site(payload) or _extract_site(nested)
    if not site and target_domain:
        site = target_domain
        hydrate_meta["site_source"] = "target_snapshot"

    if target_ticker and not payload.get("ticker"):
        payload["ticker"] = target_ticker
    if target_country and not payload.get("hq_country"):
        payload["hq_country"] = target_country
        hydrate_meta["country_source"] = "target_snapshot"
    # Fallback: use candidate.countries if present when hq_country is missing
    if not payload.get("hq_country") and isinstance(nested, dict):
        countries = nested.get("countries")
        if isinstance(countries, list) and countries:
            first = countries[0]
            if isinstance(first, dict):
                candidate_country = first.get("label") or first.get("id")
            else:
                candidate_country = str(first)
            if candidate_country:
                payload["hq_country"] = candidate_country
                hydrate_meta["country_source"] = "candidate_countries"

    # 2. SEC / EDGAR lookup to hydrate ticker, CIK, and canonical website.
    sec_record = None
    cik_hint = payload.get("cik") or (
        nested.get("cik") if isinstance(nested, dict) else None
    )
    ticker_hint = payload.get("ticker") or (
        nested.get("ticker") if isinstance(nested, dict) else None
    )
    company_name = target_snapshot.get("company_name") if target_snapshot else None
    if _table_exists(session, "sec_companies"):
        sec_record = _lookup_sec_company(
            session,
            cik=cik_hint,
            ticker=ticker_hint,
            qid=candidate.graph_entity_id,
            company_name=company_name,
        )
    if sec_record:
        match_type = sec_record.pop("_match_type", None)
        hydrate_meta["sec_match_type"] = match_type
        payload.setdefault("ticker", sec_record.get("ticker"))
        payload.setdefault("cik", sec_record.get("cik"))
        if sec_record.get("country") and not payload.get("hq_country"):
            payload["hq_country"] = sec_record.get("country")
            hydrate_meta["country_source"] = hydrate_meta.get("country_source") or "sec"
        if not site and sec_record.get("website"):
            site = sec_record.get("website")
            hydrate_meta["site_source"] = hydrate_meta.get("site_source") or "sec"
        if match_type == "qid":
            payload.setdefault("qid_to_cik_match", True)
        if sec_record.get("ticker"):
            payload.setdefault("sec_ticker", sec_record.get("ticker"))
        if sec_record.get("cik"):
            payload.setdefault("sec_cik", sec_record.get("cik"))
        if match_type:
            payload.setdefault("sec_match_type", match_type)

    # 3. Wikidata P856 fallback if still missing a site.
    if (
        not site
        and candidate.graph_source == "wikidata"
        and _table_exists(session, "wikidata_entities")
    ):
        wikidata_site = _lookup_wikidata_site(session, candidate.graph_entity_id)
        if wikidata_site:
            site = wikidata_site
            hydrate_meta["site_source"] = hydrate_meta.get("site_source") or "wikidata"

    # 3b. Populate country from Wikidata if missing.
    if (
        not payload.get("hq_country")
        and candidate.graph_source == "wikidata"
        and _table_exists(session, "wikidata_entities")
    ):
        wikidata_country = _lookup_wikidata_country(session, candidate.graph_entity_id)
        if wikidata_country:
            payload["hq_country"] = wikidata_country
            hydrate_meta["country_source"] = (
                hydrate_meta.get("country_source") or "wikidata"
            )

    # 4. Optional DBpedia fallback if everything else failed.
    if not site:
        dbpedia_site = _lookup_dbpedia_site(session, candidate.graph_entity_id)
        if dbpedia_site:
            site = dbpedia_site
            hydrate_meta["site_source"] = hydrate_meta.get("site_source") or "dbpedia"

    # 5. Store hydrated values back onto the payload (both top-level and nested candidate).
    if site:
        canonical = to_domain(site) or site
        payload["official_website"] = canonical
        if isinstance(nested, dict):
            nested.setdefault("official_website", canonical)
            nested.setdefault("domain", canonical)
            domains = nested.get("domains")
            if isinstance(domains, list):
                if canonical not in domains:
                    domains.append(canonical)
            else:
                nested["domains"] = [canonical]
            payload["candidate"] = nested

    if isinstance(nested, dict):
        ticker_value = payload.get("ticker")
        if isinstance(ticker_value, str) and ticker_value.strip():
            ticker_norm = ticker_value.strip().upper()
            tickers = nested.get("tickers")
            if isinstance(tickers, list):
                existing = {str(t).strip().upper() for t in tickers}
                if ticker_norm not in existing:
                    tickers.append(ticker_norm)
            else:
                nested["tickers"] = [ticker_norm]
        if payload.get("cik") and not nested.get("cik"):
            nested["cik"] = payload["cik"]
        country_value = payload.get("hq_country")
        if isinstance(country_value, str) and country_value.strip():
            country_norm = country_value.strip().upper()
            countries = nested.get("countries")
            if isinstance(countries, list):
                existing_codes = {
                    str(item.get("id") or item.get("label") or "").strip().upper()
                    for item in countries
                    if isinstance(item, dict)
                }
                if country_norm and country_norm not in existing_codes:
                    countries.append({"id": country_norm, "label": country_norm})
            else:
                nested["countries"] = [{"id": country_norm, "label": country_norm}]
        payload["candidate"] = nested

    if not site:
        hydrate_meta["miss_reasons"].append("missing_domain")
        if not target_domain:
            hydrate_meta["miss_reasons"].append("no_target_domain")
        if not sec_record:
            hydrate_meta["miss_reasons"].append("sec_lookup_miss")
        if candidate.graph_source != "wikidata":
            hydrate_meta["miss_reasons"].append("non_wikidata_source")
    if not payload.get("hq_country"):
        hydrate_meta["miss_reasons"].append("missing_country")

    payload["_hydrate_meta"] = hydrate_meta
    return payload


def _extract_site(payload: Dict[str, Any]) -> Optional[str]:
    for key in ("official_website", "website", "homepage", "url"):
        value = payload.get(key)
        if isinstance(value, str) and value.strip():
            return value.strip()
    return None


def _lookup_sec_company(
    session: Session,
    *,
    cik: Optional[str],
    ticker: Optional[str],
    qid: Optional[str],
    company_name: Optional[str],
) -> Optional[Dict[str, Any]]:
    """Attempt to fetch SEC metadata prioritising CIK, then ticker, QID, and finally company name."""
    queries = []
    if cik:
        queries.append(
            (
                "SELECT cik, ticker, website, country FROM sec_companies WHERE cik = :cik LIMIT 1",
                {"cik": cik},
                "cik",
            )
        )
    if ticker:
        queries.append(
            (
                "SELECT cik, ticker, website, country FROM sec_companies WHERE UPPER(ticker) = :ticker LIMIT 1",
                {"ticker": ticker.upper()},
                "ticker",
            )
        )
    if qid:
        queries.append(
            (
                "SELECT cik, ticker, website, country FROM sec_companies WHERE wikidata_qid = :qid LIMIT 1",
                {"qid": qid},
                "qid",
            )
        )
    if company_name:
        queries.append(
            (
                "SELECT cik, ticker, website, country FROM sec_companies WHERE UPPER(name) = :name LIMIT 1",
                {"name": company_name.upper()},
                "name",
            )
        )

    for sql, params, match_type in queries:
        cache_key = ("sec", match_type, tuple(sorted(params.items())))

        def loader(s=sql, p=params):
            row = _safe_first(session, s, p)
            return dict(row._mapping) if row else None

        record = _cached_lookup(cache_key, loader)
        if record:
            record["_match_type"] = match_type
            return record
    return None


def _lookup_wikidata_site(session: Session, qid: Optional[str]) -> Optional[str]:
    if not qid:
        return None
    cache_key = ("wikidata_site", (("qid", qid),))

    def loader():
        row = _safe_first(
            session,
            "SELECT official_website FROM wikidata_entities WHERE qid = :qid LIMIT 1",
            {"qid": qid},
        )
        if row and row._mapping.get("official_website"):
            return row._mapping["official_website"]
        return None

    return _cached_lookup(cache_key, loader)


def _lookup_wikidata_country(session: Session, qid: Optional[str]) -> Optional[str]:
    if not qid:
        return None
    cache_key = ("wikidata_country", (("qid", qid),))

    def loader():
        row = _safe_first(
            session,
            """
            SELECT country_code, hq_country_code
            FROM wikidata_entities
            WHERE qid = :qid
            LIMIT 1
            """,
            {"qid": qid},
        )
        if not row:
            return None
        mapping = row._mapping
        for key in ("hq_country_code", "country_code"):
            value = mapping.get(key)
            if isinstance(value, str) and value.strip():
                return value.strip().upper()
        return None

    return _cached_lookup(cache_key, loader)


def _lookup_dbpedia_site(session: Session, qid: Optional[str]) -> Optional[str]:
    if not qid or not _table_exists(session, "dbpedia_entities"):
        return None
    cache_key = ("dbpedia_site", (("qid", qid),))

    def loader():
        row = _safe_first(
            session,
            "SELECT homepage FROM dbpedia_entities WHERE qid = :qid LIMIT 1",
            {"qid": qid},
        )
        if row:
            homepage = row._mapping.get("homepage")
            if isinstance(homepage, str) and homepage.strip():
                return homepage.strip()
        return None

    return _cached_lookup(cache_key, loader)


def _safe_first(session: Session, sql: str, params: Dict[str, Any]) -> Optional[Any]:
    try:
        with _LOOKUP_SEMAPHORE:
            result = session.execute(text(sql), params)
            return result.first()
    except Exception as exc:  # pragma: no cover - diagnostics only
        log.debug("Hydrator query failed for %s with %s", sql, exc)
        session.rollback()
        return None


def _table_exists(session: Session, table_name: str) -> bool:
    cached = _TABLE_CACHE.get(table_name)
    if cached is not None:
        return cached

    try:
        result = session.execute(
            text(
                """
                SELECT EXISTS (
                  SELECT FROM information_schema.tables
                  WHERE table_name = :table_name
                )
                """
            ),
            {"table_name": table_name},
        )
        exists = bool(result.scalar() or False)
        _TABLE_CACHE[table_name] = exists
        return exists
    except Exception as exc:  # pragma: no cover - diagnostics only
        log.debug("Unable to determine presence of %s: %s", table_name, exc)
        session.rollback()
        _TABLE_CACHE[table_name] = False
        return False


def _cached_lookup(key: Tuple[str, Tuple[Tuple[str, Any], ...]], loader) -> Any:
    now = time.time()
    with _CACHE_LOCK:
        cached = _LOOKUP_CACHE.get(key)
        if cached:
            ts, value = cached
            if now - ts < _LOOKUP_CACHE_TTL:
                return value
            _LOOKUP_CACHE.pop(key, None)

    value = loader()

    with _CACHE_LOCK:
        _LOOKUP_CACHE[key] = (now, value)
        if len(_LOOKUP_CACHE) > 2048:
            stale_keys = [
                cache_key
                for cache_key, (ts, _) in _LOOKUP_CACHE.items()
                if now - ts >= _LOOKUP_CACHE_TTL
            ]
            for cache_key in stale_keys:
                _LOOKUP_CACHE.pop(cache_key, None)
    return value



from g_gremlin.impact import (
    build_impact,
    calculate_time_saved,
    format_time_saved,
    get_gremlin_message,
)


def test_time_saved_calculation():
    stats = {"issues_fixed": 10, "formulas_inspected": 100}
    minutes = calculate_time_saved(stats)
    assert minutes == 10 * 2 + int(100 * 0.1)


def test_format_time_under_hour():
    assert format_time_saved(45) == "~45 minutes"


def test_format_time_over_hour():
    result = format_time_saved(120)
    assert "hour" in result


def test_build_impact_includes_fields():
    impact = build_impact(
        action="Test action",
        stats={"issues_fixed": 5},
        scenario="lint_fixed",
        trophy="🔧",
        fixed=5,
        remaining=0,
    )
    assert "action" in impact and "stats" in impact and "gremlin_says" in impact and "trophy" in impact


def test_gremlin_message_formats():
    msg = get_gremlin_message("lint_found", count=100, issues=5)
    assert any(substr in msg for substr in ["100", "5"])



"""
Formula linting utilities for g-gremlin.
"""

from __future__ import annotations

import csv
import re
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple

from g_gremlin.graph import parse_formula_references
from g_gremlin.a1 import col_index_to_letter

# Map of issue type -> (severity, fixable)
ERROR_TYPES: dict[str, tuple[str, bool]] = {
    "unquoted_sheet_ref": ("error", True),
    "ref_error": ("error", False),
    "name_error": ("error", False),
    "value_error": ("error", False),
    "div_zero_error": ("error", False),
    "nonexistent_sheet_ref": ("error", False),
    "offset_wrong_args": ("error", False),
    "circular_ref": ("error", False),
    "iferror_missing_fallback": ("warning", False),
    "mixed_reference_style": ("warning", False),
    "hardcoded_in_formula": ("warning", False),
    "very_long_formula": ("warning", False),
}

# Human-friendly rule aliases for CLI
RULE_ALIASES: dict[str, list[str]] = {
    "no-hardcoded": ["hardcoded_in_formula"],
    "no-hardcoded-rows": ["hardcoded_in_formula"],  # alias
    "valid-refs": ["ref_error", "name_error", "nonexistent_sheet_ref"],
    "no-circular": ["circular_ref"],
    "no-errors": ["ref_error", "name_error", "value_error", "div_zero_error"],
    "dead-refs": ["nonexistent_sheet_ref"],  # User-friendly: catch broken sheet references
    "errors": ["ref_error", "name_error", "value_error", "div_zero_error"],  # Alias for no-errors
    "offset-check": ["offset_wrong_args"],
    "quoting": ["unquoted_sheet_ref"],
    "iferror-check": ["iferror_missing_fallback"],
    "mixed-refs": ["mixed_reference_style"],
    "long-formulas": ["very_long_formula"],
    "all": list(ERROR_TYPES.keys()),
}


def expand_rule_aliases(rules: List[str]) -> set[str]:
    """Expand human-friendly rule names to internal error types."""
    expanded: set[str] = set()
    for rule in rules:
        rule_lower = rule.lower().strip()
        if rule_lower in RULE_ALIASES:
            expanded.update(RULE_ALIASES[rule_lower])
        elif rule_lower in ERROR_TYPES:
            expanded.add(rule_lower)
    return expanded


def check_unquoted_sheet_refs(formula: str) -> List[Dict[str, Any]]:
    """
    Find sheet references that need quoting but are not quoted.
    """
    issues: List[Dict[str, Any]] = []
    pattern = r"(?<!')([A-Za-z0-9_][A-Za-z0-9_ \-]*[A-Za-z0-9_])!"
    for match in re.finditer(pattern, formula):
        sheet_name = match.group(1)
        if " " in sheet_name or "-" in sheet_name:
            fixed = formula.replace(f"{sheet_name}!", f"'{sheet_name}'!")
            issues.append(
                {
                    "type": "unquoted_sheet_ref",
                    "sheet_name": sheet_name,
                    "message": f"Sheet name '{sheet_name}' contains spaces/hyphens and must be quoted",
                    "formula": formula,
                    "suggested_fix": fixed,
                }
            )
    return issues


def check_error_values(value: Any) -> Dict[str, Any] | None:
    """
    Return an error issue if value is an error string.
    """
    if not isinstance(value, str):
        return None
    error_map = {
        "#REF!": "ref_error",
        "#NAME?": "name_error",
        "#VALUE!": "value_error",
        "#DIV/0!": "div_zero_error",
    }
    for token, issue_type in error_map.items():
        if token in value:
            return {
                "type": issue_type,
                "message": f"Formula returns {token} error",
                "value": value,
            }
    return None


def check_nonexistent_sheet_refs(formula: str, existing_sheets: List[str]) -> List[Dict[str, Any]]:
    """
    Detect sheet references to tabs that do not exist.
    """
    issues: List[Dict[str, Any]] = []
    existing_lower = {s.lower() for s in existing_sheets}
    # Both quoted and unquoted refs
    quoted = re.findall(r"'([^']+)'!", formula)
    unquoted = re.findall(r"(?<!')([A-Za-z0-9_][A-Za-z0-9_ \-]*[A-Za-z0-9_])!", formula)
    for ref in quoted + unquoted:
        if ref.lower() not in existing_lower:
            issues.append(
                {
                    "type": "nonexistent_sheet_ref",
                    "message": f"References sheet '{ref}' which does not exist",
                    "sheet_name": ref,
                    "formula": formula,
                }
            )
    return issues


def _split_top_level(arg_str: str) -> List[str]:
    args: List[str] = []
    current: List[str] = []
    depth = 0
    for ch in arg_str:
        if ch == "," and depth == 0:
            args.append("".join(current).strip())
            current = []
            continue
        if ch == "(":
            depth += 1
        elif ch == ")":
            depth = max(0, depth - 1)
        current.append(ch)
    if current:
        args.append("".join(current).strip())
    return args


def check_offset_args(formula: str) -> List[Dict[str, Any]]:
    """
    Ensure OFFSET calls have between 3 and 5 args.
    """
    issues: List[Dict[str, Any]] = []
    for match in re.finditer(r"OFFSET\s*\(([^)]*)\)", formula, re.IGNORECASE):
        args = _split_top_level(match.group(1))
        if len(args) < 3 or len(args) > 5:
            issues.append(
                {
                    "type": "offset_wrong_args",
                    "message": f"OFFSET requires 3-5 args, found {len(args)}",
                    "args_found": len(args),
                    "formula": formula,
                }
            )
    return issues


def check_iferror_args(formula: str) -> List[Dict[str, Any]]:
    issues: List[Dict[str, Any]] = []
    for match in re.finditer(r"IFERROR\s*\(([^)]*)\)", formula, re.IGNORECASE):
        args = _split_top_level(match.group(1))
        if len(args) < 2:
            issues.append(
                {
                    "type": "iferror_missing_fallback",
                    "message": "IFERROR has only 1 argument; add a fallback.",
                    "formula": formula,
                }
            )
    return issues


def check_mixed_reference_style(formula: str) -> Dict[str, Any] | None:
    """
    Flag formulas mixing absolute ($A$1) and relative (A1) refs.
    """
    has_abs = bool(re.search(r"\$[A-Za-z]+\$?\d+", formula))
    has_rel = bool(re.search(r"(?<!\$)[A-Za-z]{1,3}\d+", formula))
    if has_abs and has_rel:
        return {
            "type": "mixed_reference_style",
            "message": "Formula mixes absolute and relative references.",
            "formula": formula,
        }
    return None


def check_hardcoded_numbers(formula: str) -> Dict[str, Any] | None:
    """
    Rough check for magic numbers in formulas.
    """
    if re.search(r"[=+\-*/(,]\s*-?\d+(\.\d+)?", formula):
        return {
            "type": "hardcoded_in_formula",
            "message": "Formula contains hardcoded numeric constants.",
            "formula": formula,
        }
    return None


def check_formula_length(formula: str) -> Dict[str, Any] | None:
    if len(formula) > 500:
        return {
            "type": "very_long_formula",
            "message": f"Formula is {len(formula)} characters (maintainability risk).",
            "length": len(formula),
            "formula": formula[:120] + "...",
        }
    return None


def check_circular_ref(cell: str, formula: str) -> Dict[str, Any] | None:
    """
    Direct self-reference detection.
    """
    refs = parse_formula_references(formula)
    cell_core = cell.split("!", 1)[-1]
    for ref in refs:
        ref_core = ref.split("!", 1)[-1]
        if ref_core == cell_core:
            return {
                "type": "circular_ref",
                "message": "Formula references its own cell (direct circular reference).",
                "formula": formula,
            }
    return None


def lint_formula(cell: str, formula: str, value: Any, existing_sheets: List[str]) -> List[Dict[str, Any]]:
    issues: List[Dict[str, Any]] = []

    error_issue = check_error_values(value)
    if error_issue:
        error_issue["cell"] = cell
        sev, fix = ERROR_TYPES.get(error_issue["type"], ("error", False))
        error_issue["severity"] = sev
        error_issue["fixable"] = fix
        issues.append(error_issue)

    if not (isinstance(formula, str) and formula.startswith("=")):
        return issues

    for func in (
        check_unquoted_sheet_refs,
        check_nonexistent_sheet_refs,
        check_offset_args,
        check_iferror_args,
    ):
        for issue in func(formula) if func is check_unquoted_sheet_refs or func is check_offset_args or func is check_iferror_args else func(formula, existing_sheets):  # type: ignore[arg-type]
            issue["cell"] = cell
            sev, fix = ERROR_TYPES.get(issue["type"], ("error", False))
            issue["severity"] = sev
            issue["fixable"] = fix
            issues.append(issue)

    circ = check_circular_ref(cell, formula)
    if circ:
        circ["cell"] = cell
        sev, fix = ERROR_TYPES.get(circ["type"], ("error", False))
        circ["severity"] = sev
        circ["fixable"] = fix
        issues.append(circ)

    mixed = check_mixed_reference_style(formula)
    if mixed:
        mixed["cell"] = cell
        sev, fix = ERROR_TYPES.get(mixed["type"], ("warning", False))
        mixed["severity"] = sev
        mixed["fixable"] = fix
        issues.append(mixed)

    hardcoded = check_hardcoded_numbers(formula)
    if hardcoded:
        hardcoded["cell"] = cell
        sev, fix = ERROR_TYPES.get(hardcoded["type"], ("warning", False))
        hardcoded["severity"] = sev
        hardcoded["fixable"] = fix
        issues.append(hardcoded)

    long_formula = check_formula_length(formula)
    if long_formula:
        long_formula["cell"] = cell
        sev, fix = ERROR_TYPES.get(long_formula["type"], ("warning", False))
        long_formula["severity"] = sev
        long_formula["fixable"] = fix
        issues.append(long_formula)

    return issues


def generate_lint_report(issues: List[Dict[str, Any]]) -> Dict[str, Any]:
    errors = [i for i in issues if i.get("severity") == "error"]
    warnings = [i for i in issues if i.get("severity") == "warning"]
    fixable = [i for i in issues if i.get("fixable")]
    return {
        "lint_results": {
            "summary": {
                "total_issues": len(issues),
                "errors": len(errors),
                "warnings": len(warnings),
                "fixable": len(fixable),
            },
            "issues": issues,
            "fixes_available": [
                {
                    "cell": i.get("cell"),
                    "type": i.get("type"),
                    "original": i.get("formula"),
                    "fixed": i.get("suggested_fix"),
                }
                for i in fixable
                if i.get("suggested_fix")
            ],
        }
    }


def validate_csv_file(csv_path: Path, allowed_refs: Optional[List[str]] = None) -> Dict[str, Any]:
    """
    Pre-flight validate a CSV for formulas: count formulas, cross-sheet refs, and common issues.
    """
    issues: List[Dict[str, Any]] = []
    formula_count = 0
    cross_sheet_refs = 0
    allowed_lower = {r.strip().lower() for r in (allowed_refs or []) if r.strip()}

    with csv_path.open("r", newline="", encoding="utf-8-sig") as handle:
        reader = csv.reader(handle)
        rows = list(reader)

    for r_idx, row in enumerate(rows):
        for c_idx, cell in enumerate(row):
            if not isinstance(cell, str):
                continue
            a1 = f"{col_index_to_letter(c_idx + 1)}{r_idx + 1}"

            err_issue = check_error_values(cell)
            if err_issue:
                err_issue.update({"cell": a1, "severity": "error", "fixable": False})
                issues.append(err_issue)

            if not cell.startswith("="):
                continue
            formula_count += 1

            for issue in check_unquoted_sheet_refs(cell):
                issue.update({"cell": a1, "severity": "error", "fixable": True})
                issues.append(issue)

            refs = parse_formula_references(cell)
            for ref in refs:
                if "!" in ref:
                    cross_sheet_refs += 1
                    sheet_name = ref.split("!", 1)[0].strip("'")
                    if allowed_lower and sheet_name.lower() not in allowed_lower:
                        issues.append(
                            {
                                "cell": a1,
                                "type": "nonexistent_sheet_ref",
                                "severity": "error",
                                "fixable": False,
                                "message": f"References sheet '{sheet_name}' which is not in allowed refs.",
                                "sheet_name": sheet_name,
                                "formula": cell,
                            }
                        )

    errors = [i for i in issues if i.get("severity") == "error"]
    warnings = [i for i in issues if i.get("severity") == "warning"]
    fixable = [i for i in issues if i.get("fixable")]

    return {
        "summary": {
            "total_formulas": formula_count,
            "cross_sheet_refs": cross_sheet_refs,
            "errors": len(errors),
            "warnings": len(warnings),
            "fixable": len(fixable),
        },
        "issues": issues,
    }


def check_bands(csv_path: Path, band_specs: List[Tuple[int, int, str]]) -> Dict[str, Any]:
    """
    Verify that formulas within column bands reference expected sheet names.
    band_specs: list of (start_col_idx, end_col_idx, expected_sheet)
    """
    with csv_path.open("r", newline="", encoding="utf-8-sig") as handle:
        reader = csv.reader(handle)
        rows = list(reader)

    results: List[Dict[str, Any]] = []
    overall_mismatches: List[Dict[str, Any]] = []

    for start_idx, end_idx, expected in band_specs:
        formulas_checked = 0
        mismatches: List[Dict[str, Any]] = []
        for r_idx, row in enumerate(rows):
            for c_idx in range(start_idx, end_idx + 1):
                if c_idx >= len(row):
                    continue
                cell = row[c_idx]
                if not (isinstance(cell, str) and cell.startswith("=")):
                    continue
                formulas_checked += 1
                refs = parse_formula_references(cell)
                for ref in refs:
                    if "!" not in ref:
                        continue
                    sheet_name = ref.split("!", 1)[0].strip("'")
                    if sheet_name != expected:
                        a1 = f"{col_index_to_letter(c_idx + 1)}{r_idx + 1}"
                        mismatch = {
                            "cell": a1,
                            "expected": expected,
                            "found": sheet_name,
                            "formula": cell,
                        }
                        mismatches.append(mismatch)
                        overall_mismatches.append(
                            {"band": f"{col_index_to_letter(start_idx+1)}:{col_index_to_letter(end_idx+1)}", **mismatch}
                        )
        results.append(
            {
                "band": f"{col_index_to_letter(start_idx+1)}:{col_index_to_letter(end_idx+1)}",
                "expected_sheet": expected,
                "formulas_checked": formulas_checked,
                "mismatches": mismatches,
            }
        )

    return {"results": results, "mismatches": overall_mismatches}


def generate_lag_row(source_range: str, lag: int, labels: Optional[List[str]] = None) -> List[str]:
    """
    Generate a single CSV row of IFERROR(OFFSET(...)) formulas across a source range.
    """
    if ":" not in source_range:
        raise ValueError("source range must be like E5:Q5")
    start, end = source_range.split(":", 1)
    start_col = _col_letters(start)
    end_col = _col_letters(end)
    start_row = _row_number(start)
    end_row = _row_number(end)
    if start_row != end_row:
        raise ValueError("Lag generator supports single-row ranges.")
    start_idx = _letters_to_index(start_col) - 1
    end_idx = _letters_to_index(end_col) - 1
    row = labels[:] if labels else []
    for idx in range(start_idx, end_idx + 1):
        col_letter = col_index_to_letter(idx + 1)
        row.append(f"=IFERROR(OFFSET({col_letter}{start_row},0,-{lag}),0)")
    return row


def _letters_to_index(letters: str) -> int:
    value = 0
    for ch in letters.upper():
        value = value * 26 + (ord(ch) - ord("A") + 1)
    return value


def _col_letters(a1: str) -> str:
    return re.match(r"([A-Za-z]+)", a1).group(1) if re.match(r"([A-Za-z]+)", a1) else ""


def _row_number(a1: str) -> int:
    m = re.match(r"[A-Za-z]+(\d+)", a1)
    if not m:
        return 0
    return int(m.group(1))


def lint_formula_filtered(
    cell: str,
    formula: str,
    value: Any,
    existing_sheets: List[str],
    enabled_rules: Optional[set[str]] = None,
) -> List[Dict[str, Any]]:
    """
    Run lint checks but filter to only enabled rule types.
    If enabled_rules is None, run all checks.
    """
    all_issues = lint_formula(cell, formula, value, existing_sheets)
    if enabled_rules is None:
        return all_issues
    return [i for i in all_issues if i.get("type") in enabled_rules]


def detect_transitive_circular_refs(
    cells: Dict[str, Dict[str, Any]],
    default_sheet: Optional[str] = None,
) -> List[Dict[str, Any]]:
    """
    Detect transitive cycles in the dependency graph using DFS with coloring.

    Args:
        cells: Dict mapping cell A1 -> {formula, value, ...}
        default_sheet: Optional sheet name for normalizing refs

    Returns:
        List of cycle issues with cells_in_cycle
    """
    from g_gremlin.graph import build_dependency_graph

    # Build the dependency graph
    graph = build_dependency_graph(cells, default_sheet or "")

    # DFS with coloring: WHITE (0) = unvisited, GRAY (1) = in stack, BLACK (2) = done
    color: Dict[str, int] = {cell: 0 for cell in graph}
    cycles: List[List[str]] = []

    def dfs(cell: str, path: List[str]) -> None:
        if color.get(cell, 0) == 1:  # GRAY - found cycle
            cycle_start = path.index(cell) if cell in path else 0
            cycle = path[cycle_start:] + [cell]
            cycles.append(cycle)
            return
        if color.get(cell, 0) == 2:  # BLACK - already processed
            return

        color[cell] = 1  # GRAY
        path.append(cell)

        for ref in graph.get(cell, {}).get("references", []):
            if ref in graph:
                dfs(ref, path)

        path.pop()
        color[cell] = 2  # BLACK

    for cell in graph:
        if color.get(cell, 0) == 0:
            dfs(cell, [])

    # Convert cycles to issues
    issues: List[Dict[str, Any]] = []
    seen_cycles: set[tuple[str, ...]] = set()

    for cycle in cycles:
        # Normalize cycle to avoid duplicates (start from smallest element)
        if len(cycle) > 1:
            min_idx = cycle.index(min(cycle[:-1]))
            normalized = tuple(cycle[min_idx:-1])
            if normalized not in seen_cycles:
                seen_cycles.add(normalized)
                issues.append({
                    "type": "circular_ref",
                    "severity": "error",
                    "fixable": False,
                    "message": f"Transitive circular reference detected: {' -> '.join(cycle)}",
                    "cells_in_cycle": list(cycle),
                    "cycle_length": len(cycle) - 1,
                })

    return issues



"""
Admin endpoint for synchronising tier changes from Stripe/Clerk.

The web application (Next.js) calls this endpoint after updating Clerk metadata
so that the FastAPI backend maintains the same subscription tier and status.
"""

from __future__ import annotations

import logging
from datetime import datetime
from typing import Literal, Optional

from fastapi import APIRouter, Depends, Header, HTTPException
from pydantic import BaseModel, root_validator
from sqlalchemy import select
from sqlalchemy.ext.asyncio import AsyncSession

from ... import settings as saas_settings
from ...db import get_session
from ...middleware.quota_checker import TIER_LIMITS
from ...models import UsageQuotaModel
from ...services.usage_gateway import UsageGateway

log = logging.getLogger("fmatch.api.sync_tier")

router = APIRouter(prefix="/api/v1/admin", tags=["admin-internal"])


class TierSyncPayload(BaseModel):
    """Payload for tier synchronisation from Stripe webhooks or admin tooling."""

    clerk_user_id: str
    tier: Literal["free", "pro", "scale", "unleashed"]
    status: str  # active, canceled, past_due, trialing, etc.
    monthly_enrichments: Optional[int] = None
    monthly_match_limit: Optional[int] = None
    period_start: Optional[str] = None
    period_end: Optional[str] = None
    source: Optional[str] = "webhook"
    force_reset: Optional[bool] = (
        False  # Allow admin scripts to bypass downgrade protection
    )
    addons: Optional[list[str]] = None

    @root_validator(pre=True)
    def _normalise_aliases(cls, values: dict[str, object]) -> dict[str, object]:
        """Allow camelCase aliases used by the web team."""
        if "monthlyEnrichmentCredits" in values and "monthly_enrichments" not in values:
            values["monthly_enrichments"] = values["monthlyEnrichmentCredits"]
        if "monthlyMatchLimit" in values and "monthly_match_limit" not in values:
            values["monthly_match_limit"] = values["monthlyMatchLimit"]
        if "source" not in values and "event_source" in values:
            values["source"] = values["event_source"]
        return values


async def require_webhook_secret(
    authorization: str = Header(..., alias="Authorization"),
) -> str:
    """
    Verify webhook secret for internal service-to-service calls.

    Expected format: "Bearer <WEBHOOK_SECRET>"
    """

    if not authorization.startswith("Bearer "):
        raise HTTPException(
            status_code=401, detail="Invalid authorization header format"
        )

    token = authorization.replace("Bearer ", "")

    # Development escape hatch to simplify local testing.
    if saas_settings.ENV.lower() in ("dev", "development") and token == "dev-token":
        log.info("[sync-tier] Allowing dev-token in development")
        return token

    webhook_secret = saas_settings.WEBHOOK_SECRET
    if not webhook_secret:
        log.error("[sync-tier] WEBHOOK_SECRET not configured!")
        raise HTTPException(
            status_code=500, detail="Webhook authentication not configured"
        )

    if token != webhook_secret:
        log.warning("[sync-tier] Invalid webhook secret provided")
        raise HTTPException(status_code=401, detail="Invalid webhook secret")

    return token


def _resolve_entitlements(payload: TierSyncPayload) -> tuple[int, int]:
    """Determine entitlement values using payload overrides or tier defaults."""
    tier_limits = TIER_LIMITS.get(payload.tier, TIER_LIMITS["free"])
    monthly_rows = payload.monthly_match_limit or tier_limits["monthly_rows"]
    monthly_enrichments = (
        payload.monthly_enrichments or tier_limits["enrichment_credits"]
    )
    return monthly_rows, monthly_enrichments


@router.post("/sync-tier")
async def sync_tier_from_stripe(
    payload: TierSyncPayload,
    _auth: str = Depends(require_webhook_secret),
    db: AsyncSession = Depends(get_session),
) -> dict:
    """
    Update UsageQuotaModel to reflect the tier coming from Stripe/Clerk.

    Flow:
        1. Stripe fires webhook to Next.js
        2. Next.js updates PostgreSQL & Clerk metadata
        3. Next.js calls this endpoint with the same tier information
    """

    tier_limits = TIER_LIMITS.get(payload.tier)
    if not tier_limits:
        raise HTTPException(
            status_code=400,
            detail=f"Invalid tier: {payload.tier}. Must be one of: {list(TIER_LIMITS.keys())}",
        )

    monthly_rows, monthly_enrichments = _resolve_entitlements(payload)

    enrichments_str = (
        f", enrichments={monthly_enrichments}"
        if monthly_enrichments is not None
        else ""
    )
    log.info(
        "[sync-tier] Received tier sync for user %s: tier=%s, status=%s%s, source=%s",
        payload.clerk_user_id,
        payload.tier,
        payload.status,
        enrichments_str,
        payload.source,
    )

    stmt = select(UsageQuotaModel).where(
        UsageQuotaModel.account_id == payload.clerk_user_id
    )
    result = await db.execute(stmt)
    quota = result.scalar_one_or_none()

    now = datetime.utcnow()

    if not quota:
        quota = UsageQuotaModel(
            account_id=payload.clerk_user_id,
            tier=payload.tier,
            monthly_rows_used=0,
            enrichment_credits_used=0,
            subscription_status=payload.status or "unknown",
            tier_synced_at=now,
            tier_sync_source=payload.source or "webhook",
            monthly_enrichment_entitlement=monthly_enrichments,
            monthly_match_entitlement=monthly_rows,
        )
        db.add(quota)
        await db.commit()
        await db.refresh(quota)

        try:
            gateway = UsageGateway(db)
            await gateway.sync_from_stripe(
                account_id=payload.clerk_user_id,
                tier=payload.tier,
                addons=payload.addons or [],
                source=payload.source or "webhook",
                monthly_row_limit=monthly_rows,
            )
        except Exception as exc:  # noqa: BLE001 - best effort
            log.warning("[sync-tier] Failed to sync usage gateway: %s", exc)

        return {
            "ok": True,
            "updated": True,
            "message": f"Created new quota record with tier={payload.tier}",
            "account_id": payload.clerk_user_id,
            "tier": quota.tier,
            "status": quota.subscription_status,
            "enrichments": quota.monthly_enrichment_entitlement,
            "monthly_match_limit": quota.monthly_match_entitlement,
            "source": quota.tier_sync_source,
        }

    old_tier = quota.tier

    # Protect against erroneous downgrades from paid to free
    # If current tier is paid (pro/scale) and new tier is free, check if this is legitimate
    if (
        old_tier in ("pro", "scale")
        and payload.tier == "free"
        and not payload.force_reset
    ):
        # Check if there's an active subscription in the database
        from sqlalchemy import text

        active_sub = await db.execute(
            text("""
                SELECT s.stripe_subscription_id, s.status
                FROM subscription s
                WHERE s.account_id = :aid
                AND s.status IN ('active', 'trialing')
                ORDER BY s.created_at DESC
                LIMIT 1
            """),
            {"aid": payload.clerk_user_id},
        )
        active_subscription = active_sub.first()

        if active_subscription:
            # There's an active subscription - reject the downgrade regardless of status
            log.warning(
                "[sync-tier] Rejecting downgrade from %s to free - active subscription %s exists (status=%s). "
                "If this is legitimate, cancel the subscription first or use force_reset=true. Source: %s",
                old_tier,
                active_subscription[0],
                active_subscription[1],
                payload.source,
            )
            return {
                "ok": True,
                "updated": False,
                "message": f"Rejected downgrade from {old_tier} to free - active subscription exists",
                "account_id": payload.clerk_user_id,
                "old_tier": old_tier,
                "new_tier": old_tier,  # Kept old tier
                "status": quota.subscription_status,
                "enrichments": quota.monthly_enrichment_entitlement,
                "monthly_match_limit": quota.monthly_match_entitlement,
                "source": quota.tier_sync_source,
                "active_subscription": active_subscription[0],
            }

    # Log if force_reset was used
    if payload.force_reset and old_tier in ("pro", "scale") and payload.tier == "free":
        log.info(
            "[sync-tier] Force reset enabled - allowing downgrade from %s to free. Source: %s",
            old_tier,
            payload.source,
        )

    quota.tier = payload.tier
    quota.subscription_status = payload.status or quota.subscription_status or "unknown"
    quota.tier_synced_at = now
    quota.tier_sync_source = payload.source or quota.tier_sync_source or "webhook"
    quota.monthly_enrichment_entitlement = monthly_enrichments
    quota.monthly_match_entitlement = monthly_rows

    await db.commit()
    await db.refresh(quota)

    try:
        gateway = UsageGateway(db)
        await gateway.sync_from_stripe(
            account_id=payload.clerk_user_id,
            tier=payload.tier,
            addons=payload.addons or [],
            source=payload.source or "webhook",
            monthly_row_limit=monthly_rows,
        )
    except Exception as exc:  # noqa: BLE001 - best effort
        log.warning("[sync-tier] Failed to sync usage gateway: %s", exc)

    log.info(
        "[sync-tier] Updated tier for %s: %s -> %s (status=%s, source=%s)",
        payload.clerk_user_id,
        old_tier,
        quota.tier,
        quota.subscription_status,
        quota.tier_sync_source,
    )

    return {
        "ok": True,
        "updated": old_tier != quota.tier
        or quota.subscription_status != payload.status
        or quota.monthly_enrichment_entitlement != monthly_enrichments,
        "message": f"Tier updated from {old_tier} to {quota.tier}",
        "account_id": payload.clerk_user_id,
        "old_tier": old_tier,
        "new_tier": quota.tier,
        "status": quota.subscription_status,
        "enrichments": quota.monthly_enrichment_entitlement,
        "monthly_match_limit": quota.monthly_match_entitlement,
        "source": quota.tier_sync_source,
    }
